import json
import logging
from typing import Any, Dict, Optional

from opentelemetry import context as context_api
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.trace import Span

logger = logging.getLogger(__name__)


def should_suppress_instrumentation() -> bool:
    """Check if instrumentation should be suppressed."""
    return context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY) is True


def set_request_attributes(span: Span, kwargs: Dict[str, Any], source_type: Optional[str] = None) -> None:
    """
    Set the request attributes for the span.

    Args:
        span: The span to set the attributes on.
        kwargs: The keyword arguments to extract the attributes from.
        source_type: The type of the source (e.g. "url" or "file").
    """
    if not span.is_recording():
        return

    GENERIC_ATTRIBUTES = [
        "callback",
        "extra",
        "model",
        "language",
        "encoding",
        "multichannel",
        "diarize",
        "detect_language",
        "detect_entities",
        "sentiment",
        "summarize",
        "topics",
        "intents",
        "tag",
        "custom_topic",
        "custom_intent",
    ]

    for key in GENERIC_ATTRIBUTES:
        value = kwargs.get(key)
        if value is not None:
            span.set_attribute(f"gen_ai.request.{key}", value)

    if source_type == "url" and "url" in kwargs:
        span.set_attribute("gen_ai.request.url", str(kwargs["url"]))

    if text := kwargs.get("text"):
        character_count = len(text)
        span.set_attribute("gen_ai.usage.prompt.character_count", character_count)
        span.set_attribute("gen_ai.prompt.0.role", "Input")
        span.set_attribute("gen_ai.prompt.0.content", str(text))

    if request := kwargs.get("request"):
        if isinstance(request, dict):
            text = request.get("text", "")
            character_count = len(text)
            span.set_attribute("gen_ai.usage.prompt.character_count", character_count)
            span.set_attribute("gen_ai.prompt.0.role", "Input")
            span.set_attribute("gen_ai.prompt.0.content", str(text))

    if source_type == "file" and "request" in kwargs:
        request_value = kwargs["request"]
        try:
            if isinstance(request_value, (bytes, bytearray)):
                span.set_attribute("gen_ai.request.size_bytes", len(request_value))
        except Exception as e:
            logger.debug("Failed to set Deepgram request size: %s", e)


def _set_websocket_results_event_attributes(span: Span, message: Any) -> None:
    """
    Handle ListenV1ResultsEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.results")

    if duration := getattr(message, "duration", None):
        span.set_attribute("deepgram.websocket.results.duration", duration)
    if start := getattr(message, "start", None):
        span.set_attribute("deepgram.websocket.results.start", start)
    if is_final := getattr(message, "is_final", None):
        span.set_attribute("deepgram.websocket.results.is_final", is_final)
    if speech_final := getattr(message, "speech_final", None):
        span.set_attribute("deepgram.websocket.results.speech_final", speech_final)
    if channel_index := getattr(message, "channel_index", None):
        span.set_attribute("deepgram.websocket.results.channel_index", channel_index)

    # Extract transcript from channel.alternatives
    if channel := getattr(message, "channel", None):
        if alternatives := getattr(channel, "alternatives", None):
            first_alt = alternatives[0] if isinstance(alternatives, list) else next(iter(alternatives), None)
            if first_alt:
                if transcript := getattr(first_alt, "transcript", None):
                    span.set_attribute("gen_ai.completion.0.role", "Transcribed Text")
                    span.set_attribute("gen_ai.completion.0.content", transcript)
                if confidence := getattr(first_alt, "confidence", None):
                    span.set_attribute("deepgram.websocket.results.confidence", confidence)
                if languages := getattr(first_alt, "languages", None):
                    span.set_attribute("deepgram.websocket.results.languages", languages)

    # Extract model info from metadata
    if metadata := getattr(message, "metadata", None):
        if request_id := getattr(metadata, "request_id", None):
            span.set_attribute("gen_ai.response.request_id", str(request_id))
        if model_uuid := getattr(metadata, "model_uuid", None):
            span.set_attribute("gen_ai.response.model_uuid", str(model_uuid))
        if model_info := getattr(metadata, "model_info", None):
            if name := getattr(model_info, "name", None):
                span.set_attribute("gen_ai.request.model", str(name))
            if version := getattr(model_info, "version", None):
                span.set_attribute("gen_ai.request.model_version", str(version))
            if arch := getattr(model_info, "arch", None):
                span.set_attribute("gen_ai.request.model_arch", str(arch))


def _set_websocket_metadata_event_attributes(span: Span, message: Any) -> None:
    """
    Handle ListenV1MetadataEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.metadata")

    if request_id := getattr(message, "request_id", None):
        span.set_attribute("gen_ai.response.request_id", str(request_id))
    if created := getattr(message, "created", None):
        span.set_attribute("deepgram.websocket.metadata.created", str(created))
    if duration := getattr(message, "duration", None):
        span.set_attribute("gen_ai.audio.duration", duration / 60)
    if channels := getattr(message, "channels", None):
        span.set_attribute("gen_ai.response.channels", int(channels))


def _set_websocket_utterance_end_event_attributes(span: Span, message: Any) -> None:
    """
    Handle ListenV1UtteranceEndEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.utterance_end")

    if channel := getattr(message, "channel", None):
        span.set_attribute("deepgram.websocket.utterance_end.channel", channel)
    if last_word_end := getattr(message, "last_word_end", None):
        span.set_attribute("deepgram.websocket.utterance_end.last_word_end", last_word_end)


def _set_websocket_speech_started_event_attributes(span: Span, message: Any) -> None:
    """
    Handle ListenV1SpeechStartedEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.speech_started")

    if channel := getattr(message, "channel", None):
        span.set_attribute("deepgram.websocket.speech_started.channel", channel)
    if timestamp := getattr(message, "timestamp", None):
        span.set_attribute("deepgram.websocket.speech_started.timestamp", timestamp)


def _set_websocket_v2_connected_event_attributes(span: Span, message: Any) -> None:
    """
    Handle ListenV2ConnectedEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.v2.connected")

    if request_id := getattr(message, "request_id", None):
        span.set_attribute("gen_ai.response.request_id", str(request_id))
    if sequence_id := getattr(message, "sequence_id", None):
        span.set_attribute("deepgram.websocket.v2.sequence_id", sequence_id)


def _set_websocket_v2_turn_info_event_attributes(span: Span, message: Any) -> None:
    """
    Handle ListenV2TurnInfoEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.v2.turn_info")

    if request_id := getattr(message, "request_id", None):
        span.set_attribute("gen_ai.response.request_id", str(request_id))
    if sequence_id := getattr(message, "sequence_id", None):
        span.set_attribute("deepgram.websocket.v2.sequence_id", sequence_id)
    if event := getattr(message, "event", None):
        span.set_attribute("deepgram.websocket.v2.event", str(event))
    if turn_index := getattr(message, "turn_index", None):
        span.set_attribute("deepgram.websocket.v2.turn_index", turn_index)
    if audio_window_start := getattr(message, "audio_window_start", None):
        span.set_attribute("deepgram.websocket.v2.audio_window_start", audio_window_start)
    if audio_window_end := getattr(message, "audio_window_end", None):
        span.set_attribute("deepgram.websocket.v2.audio_window_end", audio_window_end)
    if transcript := getattr(message, "transcript", None):
        span.set_attribute("gen_ai.completion.0.role", "Transcribed Text")
        span.set_attribute("gen_ai.completion.0.content", transcript)
    if end_of_turn_confidence := getattr(message, "end_of_turn_confidence", None):
        span.set_attribute("deepgram.websocket.v2.end_of_turn_confidence", end_of_turn_confidence)


def _set_websocket_v2_fatal_error_event_attributes(span: Span, message: Any) -> None:
    """
    Handle ListenV2FatalErrorEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.v2.fatal_error")

    if sequence_id := getattr(message, "sequence_id", None):
        span.set_attribute("deepgram.websocket.v2.sequence_id", sequence_id)
    if code := getattr(message, "code", None):
        span.set_attribute("deepgram.websocket.v2.error.code", str(code))
    if description := getattr(message, "description", None):
        span.set_attribute("deepgram.websocket.v2.error.description", str(description))


def _set_agent_v1_welcome_attributes(span: Span, message: Any) -> None:
    """
    Handle AgentV1WelcomeMessage WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.agent.welcome")

    if request_id := getattr(message, "request_id", None):
        span.set_attribute("gen_ai.response.request_id", str(request_id))


def _set_agent_v1_history_message_attributes(span: Span, message: Any) -> None:
    """
    Handle AgentV1HistoryMessage WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.agent.history")

    if role := getattr(message, "role", None):
        span.set_attribute("deepgram.agent.history.role", str(role))
    if content := getattr(message, "content", None):
        span.set_attribute("deepgram.agent.history.content", str(content))


def _set_agent_v1_conversation_text_attributes(span: Span, message: Any) -> None:
    """
    Handle AgentV1ConversationTextEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.agent.conversation_text")

    if role := getattr(message, "role", None):
        span.set_attribute("deepgram.agent.conversation.role", str(role))
    if content := getattr(message, "content", None):
        span.set_attribute("deepgram.agent.conversation.content", str(content))


def _set_agent_v1_agent_thinking_attributes(span: Span, message: Any) -> None:
    """
    Handle AgentV1AgentThinkingEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.agent.thinking")

    if content := getattr(message, "content", None):
        span.set_attribute("deepgram.agent.thinking.content", str(content))


def _set_agent_v1_agent_started_speaking_attributes(span: Span, message: Any) -> None:
    """
    Handle AgentV1AgentStartedSpeakingEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.agent.started_speaking")

    if total_latency := getattr(message, "total_latency", None):
        span.set_attribute("deepgram.agent.latency.total", total_latency)
    if tts_latency := getattr(message, "tts_latency", None):
        span.set_attribute("deepgram.agent.latency.tts", tts_latency)
    if ttt_latency := getattr(message, "ttt_latency", None):
        span.set_attribute("deepgram.agent.latency.ttt", ttt_latency)


def _set_agent_v1_injection_refused_attributes(span: Span, message: Any) -> None:
    """
    Handle AgentV1InjectionRefusedEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.agent.injection_refused")

    if msg := getattr(message, "message", None):
        span.set_attribute("deepgram.agent.injection_refused.message", str(msg))


def _set_agent_v1_error_attributes(span: Span, message: Any) -> None:
    """
    Handle AgentV1ErrorEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.agent.error")

    if description := getattr(message, "description", None):
        span.set_attribute("deepgram.agent.error.description", str(description))
    if code := getattr(message, "code", None):
        span.set_attribute("deepgram.agent.error.code", str(code))


def _set_agent_v1_warning_attributes(span: Span, message: Any) -> None:
    """
    Handle AgentV1WarningEvent WebSocket message.

    Args:
        span: The span to set the attributes on.
        message: The message to extract the attributes from.
    """
    span.add_event("deepgram.agent.warning")

    if description := getattr(message, "description", None):
        span.set_attribute("deepgram.agent.warning.description", str(description))
    if code := getattr(message, "code", None):
        span.set_attribute("deepgram.agent.warning.code", str(code))


def _set_http_response_attributes(span: Span, response: Any) -> None:
    """
    Handle HTTP API response attributes (transcribe_url, transcribe_file, analyze, generate).

    Args:
        span: The span to set the attributes on.
        response: The response to extract the attributes from.
    """
    if results := getattr(response, "results", None):
        if channels := getattr(results, "channels", None):
            first_channel = next(iter(channels))
            if alternatives := getattr(first_channel, "alternatives", None):
                first_alt = next(iter(alternatives))
                if transcript := getattr(first_alt, "transcript", None):
                    span.set_attribute("gen_ai.completion.0.role", "Transcribed Text")
                    span.set_attribute("gen_ai.completion.0.content", transcript)

        if summary := getattr(results, "summary", None):
            if result := getattr(summary, "result", None):
                span.set_attribute("gen_ai.response.summary.result", result)
            if text := getattr(summary, "text", None):
                span.set_attribute("gen_ai.response.summary.text", text)
            if short := getattr(summary, "short", None):
                span.set_attribute("gen_ai.response.summary.short", short)

        if topics := getattr(results, "topics", None):
            if segments := getattr(topics, "segments", None):
                for i, segment in enumerate(segments):
                    span.set_attribute("gen_ai.response.topics", json.dumps(segment))

        if intents := getattr(results, "intents", None):
            if segments := getattr(intents, "segments", None):
                for i, segment in enumerate(segments):
                    span.set_attribute("gen_ai.response.intents", json.dumps(segment))

        if sentiments := getattr(results, "sentiments", None):
            if average := getattr(sentiments, "average", None):
                if sentiment := getattr(average, "sentiment", None):
                    span.set_attribute("gen_ai.response.sentiments.average", sentiment)
                if sentiment_score := getattr(average, "sentiment_score", None):
                    span.set_attribute("gen_ai.response.sentiments.average_score", sentiment_score)

    if metadata := getattr(response, "metadata", None):
        if request_id := getattr(metadata, "request_id", None):
            span.set_attribute("gen_ai.response.request_id", str(request_id))

        if duration := getattr(metadata, "duration", None):
            span.set_attribute("gen_ai.audio.duration", duration / 60)

        if channel_count := getattr(metadata, "channels", None):
            span.set_attribute("gen_ai.response.channels", channel_count)

        if models := getattr(metadata, "models", None):
            span.set_attribute("gen_ai.response.model_list", list(models) if not isinstance(models, str) else [models])

        if model_info := getattr(metadata, "model_info", None):
            if isinstance(model_info, dict) and model_info:
                first_model_key = next(iter(model_info))
                info = model_info.get(first_model_key, {})
            if name := info.get("name", None):
                span.set_attribute("gen_ai.request.model", str(name))
            if version := info.get("version", None):
                span.set_attribute("gen_ai.request.model_version", str(version))
            if arch := info.get("arch", None):
                span.set_attribute("gen_ai.request.model_arch", str(arch))

        if summary_info := getattr(metadata, "summary_info", None):
            if model_uuid := getattr(summary_info, "model_uuid", None) or summary_info.get("model_uuid", None):
                span.set_attribute("gen_ai.response.summary.model_uuid", str(model_uuid))
            if summary_input_tokens := getattr(summary_info, "input_tokens", None) or summary_info.get(
                "input_tokens", None
            ):
                span.set_attribute("gen_ai.usage.summary.prompt_tokens", str(summary_input_tokens))
            if summary_output_tokens := getattr(summary_info, "output_tokens", None) or summary_info.get(
                "output_tokens", None
            ):
                span.set_attribute("gen_ai.usage.summary.completion_tokens", str(summary_output_tokens))
        if sentiment_info := getattr(metadata, "sentiment_info", None):
            if model_uuid := getattr(sentiment_info, "model_uuid", None) or sentiment_info.get("model_uuid", None):
                span.set_attribute("gen_ai.response.sentiment.model_uuid", str(model_uuid))
            if sentiment_input_tokens := getattr(sentiment_info, "input_tokens", None) or sentiment_info.get(
                "input_tokens", None
            ):
                span.set_attribute("gen_ai.usage.sentiment.prompt_tokens", str(sentiment_input_tokens))
            if sentiment_output_tokens := getattr(sentiment_info, "output_tokens", None) or sentiment_info.get(
                "output_tokens", None
            ):
                span.set_attribute("gen_ai.usage.sentiment.completion_tokens", str(sentiment_output_tokens))

        if topics_info := getattr(metadata, "topics_info", None):
            if model_uuid := getattr(topics_info, "model_uuid", None) or topics_info.get("model_uuid", None):
                span.set_attribute("gen_ai.response.topics.model_uuid", str(model_uuid))
            if topics_input_tokens := getattr(topics_info, "input_tokens", None) or topics_info.get(
                "input_tokens", None
            ):
                span.set_attribute("gen_ai.usage.topics.prompt_tokens", str(topics_input_tokens))
            if topics_output_tokens := getattr(topics_info, "output_tokens", None) or topics_info.get(
                "output_tokens", None
            ):
                span.set_attribute("gen_ai.usage.topics.completion_tokens", str(topics_output_tokens))

        if intents_info := getattr(metadata, "intents_info", None):
            if model_uuid := getattr(intents_info, "model_uuid", None) or intents_info.get("model_uuid", None):
                span.set_attribute("gen_ai.response.intents.model_uuid", str(model_uuid))
            if intents_input_tokens := getattr(intents_info, "input_tokens", None) or intents_info.get(
                "input_tokens", None
            ):
                span.set_attribute("gen_ai.usage.intents.prompt_tokens", str(intents_input_tokens))
            if intents_output_tokens := getattr(intents_info, "output_tokens", None) or intents_info.get(
                "output_tokens", None
            ):
                span.set_attribute("gen_ai.usage.intents.completion_tokens", str(intents_output_tokens))


def set_response_attributes(span: Span, response: Any) -> None:
    """
    Set the response attributes for the span.

    Args:
        span: The span to set the attributes on.
        response: The response to extract the attributes from.
    """
    if not span.is_recording():
        return

    try:
        msg_type = getattr(response, "type", None)

        # ListenV1 WebSocket message types
        if msg_type == "Results":
            _set_websocket_results_event_attributes(span, response)
        elif msg_type == "Metadata":
            _set_websocket_metadata_event_attributes(span, response)
        elif msg_type == "UtteranceEnd":
            _set_websocket_utterance_end_event_attributes(span, response)
        elif msg_type == "SpeechStarted":
            _set_websocket_speech_started_event_attributes(span, response)

        # ListenV2 WebSocket message types
        elif msg_type == "Connected":
            _set_websocket_v2_connected_event_attributes(span, response)
        elif msg_type == "TurnInfo":
            _set_websocket_v2_turn_info_event_attributes(span, response)
        elif msg_type == "FatalError":
            _set_websocket_v2_fatal_error_event_attributes(span, response)

        # AgentV1 WebSocket message types
        elif msg_type == "Welcome":
            _set_agent_v1_welcome_attributes(span, response)
        elif msg_type == "History":
            _set_agent_v1_history_message_attributes(span, response)
        elif msg_type == "ConversationText":
            _set_agent_v1_conversation_text_attributes(span, response)
        elif msg_type == "AgentThinking":
            _set_agent_v1_agent_thinking_attributes(span, response)
        elif msg_type == "AgentStartedSpeaking":
            _set_agent_v1_agent_started_speaking_attributes(span, response)
        elif msg_type == "InjectionRefused":
            _set_agent_v1_injection_refused_attributes(span, response)
        elif msg_type == "Error":
            _set_agent_v1_error_attributes(span, response)
        elif msg_type == "Warning":
            _set_agent_v1_warning_attributes(span, response)
        else:
            _set_http_response_attributes(span, response)

        span.set_attribute("netra.span.type", "GENERATION")

    except Exception as e:
        logger.debug("Failed to set Deepgram response attributes: %s", e)



"""OpenTelemetry Google GenAI API instrumentation"""

import logging
import os
from typing import Any, Collection

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.trace import get_tracer
from wrapt import wrap_function_wrapper

from netra.instrumentation.google_genai.version import __version__
from netra.instrumentation.google_genai.wrappers import (
    acontent_stream_wrapper,
    acontent_wrapper,
    aimages_wrapper,
    avideos_wrapper,
    content_stream_wrapper,
    content_wrapper,
    images_wrapper,
    videos_wrapper,
)

logger = logging.getLogger(__name__)

_instruments = ("google-genai >= 0.1.0",)


class NetraGoogleGenAiInstrumentor(BaseInstrumentor):  # type: ignore
    """Custom Google GenAI instrumentor for Netra SDK."""

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs: Any) -> None:
        """Instrument Google GenAI client methods"""

        try:
            tracer_provider = kwargs.get("tracer_provider")
            tracer = get_tracer(__name__, __version__, tracer_provider)
        except Exception as e:
            logger.error(f"Failed to initialize tracer: {e}")
            return

        try:
            wrap_function_wrapper(
                "google.genai.models",
                "Models.generate_content",
                content_wrapper(tracer),
            )
            wrap_function_wrapper(
                "google.genai.models",
                "AsyncModels.generate_content",
                acontent_wrapper(tracer),
            )
        except Exception as e:
            logger.error(f"Failed to instrument generate_content: {e}")

        try:
            wrap_function_wrapper(
                "google.genai.models",
                "Models.generate_content_stream",
                content_stream_wrapper(tracer),
            )
            wrap_function_wrapper(
                "google.genai.models",
                "AsyncModels.generate_content_stream",
                acontent_stream_wrapper(tracer),
            )
        except Exception as e:
            logger.error(f"Failed to instrument generate_content_stream: {e}")

        try:
            wrap_function_wrapper(
                "google.genai.models",
                "Models.generate_images",
                images_wrapper(tracer),
            )
            wrap_function_wrapper(
                "google.genai.models",
                "AsyncModels.generate_images",
                aimages_wrapper(tracer),
            )
        except Exception as e:
            logger.error(f"Failed to instrument generate_images: {e}")

        try:
            wrap_function_wrapper(
                "google.genai.models",
                "Models.generate_videos",
                videos_wrapper(tracer),
            )
            wrap_function_wrapper(
                "google.genai.models",
                "AsyncModels.generate_videos",
                avideos_wrapper(tracer),
            )
        except Exception as e:
            logger.error(f"Failed to instrument generate_videos: {e}")

    def _uninstrument(self, **kwargs: Any) -> None:
        try:
            unwrap("google.genai.models", "Models.generate_content")
            unwrap("google.genai.models", "AsyncModels.generate_content")
        except (AttributeError, ModuleNotFoundError):
            logger.error("Failed to uninstrument generate_content")

        try:
            unwrap("google.genai.models", "Models.generate_content_stream")
            unwrap("google.genai.models", "AsyncModels.generate_content_stream")
        except (AttributeError, ModuleNotFoundError):
            logger.error("Failed to uninstrument generate_content_stream")

        try:
            unwrap("google.genai.models", "Models.generate_images")
            unwrap("google.genai.models", "AsyncModels.generate_images")
        except (AttributeError, ModuleNotFoundError):
            logger.error("Failed to uninstrument generate_images")

        try:
            unwrap("google.genai.models", "Models.generate_videos")
            unwrap("google.genai.models", "AsyncModels.generate_videos")
        except (AttributeError, ModuleNotFoundError):
            logger.error("Failed to uninstrument generate_videos")



import json
import os
from typing import Any, Dict, List, Optional

from opentelemetry.util.re import parse_env_headers

from netra.version import __version__


class Config:
    """
    Holds configuration options for the tracer.
    """

    # SDK Constants
    SDK_NAME = "netra"
    LIBRARY_NAME = "netra"
    LIBRARY_VERSION = __version__
    ATTRIBUTE_MAX_LEN = int(os.getenv("NETRA_ATTRIBUTE_MAX_LEN", 50000))
    CONVERSATION_MAX_LEN = int(os.getenv("NETRA_CONVERSATION_CONTENT_MAX_LEN", 50000))
    TRIAL_BLOCK_DURATION_SECONDS = int(os.getenv("TRIAL_BLOCK_DURATION_SECONDS", 15 * 60))

    def __init__(
        self,
        app_name: Optional[str] = None,
        headers: Optional[str] = None,
        disable_batch: Optional[bool] = None,
        trace_content: Optional[bool] = None,
        debug_mode: Optional[bool] = None,
        enable_root_span: Optional[bool] = None,
        resource_attributes: Optional[Dict[str, Any]] = None,
        environment: Optional[str] = None,
        enable_scrubbing: Optional[bool] = None,
        blocked_spans: Optional[List[str]] = None,
    ):
        """
        Initialize the configuration.

        Args:
            app_name: Logical name for this service
            headers: Additional headers (W3C Correlation-Context format)
            disable_batch: Whether to disable batch span processor
            trace_content: Whether to capture prompt/completion content
            debug_mode: Whether to enable SDK logging (default: False)
            enable_root_span: Whether to create a process root span (default: False)
            resource_attributes: Custom resource attributes dict (e.g., {'env': 'prod', 'version': '1.0.0'})
            enable_scrubbing: Whether to enable pydantic logfire scrubbing (default: False)
            blocked_spans: List of span names (prefix/suffix patterns) to block from export
        """
        self.app_name = self._get_app_name(app_name)
        self.otlp_endpoint = self._get_otlp_endpoint()
        self.api_key = os.getenv("NETRA_API_KEY")
        self.headers = self._parse_headers(headers)

        self._validate_api_key()
        self._setup_authentication()

        self.disable_batch = self._get_bool_config(disable_batch, "NETRA_DISABLE_BATCH", default=False)
        self.trace_content = self._get_bool_config(trace_content, "NETRA_TRACE_CONTENT", default=True)
        self.debug_mode = self._get_bool_config(debug_mode, "NETRA_DEBUG", default=False)
        self.enable_root_span = self._get_bool_config(enable_root_span, "NETRA_ENABLE_ROOT_SPAN", default=False)
        self.enable_scrubbing = self._get_bool_config(enable_scrubbing, "NETRA_ENABLE_SCRUBBING", default=False)

        self.environment = environment or os.getenv("NETRA_ENV", "default")
        self.resource_attributes = self._get_resource_attributes(resource_attributes)
        self.blocked_spans = blocked_spans

        self._set_trace_content_env()

    def _get_app_name(self, app_name: Optional[str]) -> str:
        """Get application name from param or environment variables."""
        return app_name or os.getenv("NETRA_APP_NAME") or os.getenv("OTEL_SERVICE_NAME") or "llm_tracing_service"

    def _get_otlp_endpoint(self) -> str | None:
        """Get OTLP endpoint from environment variables."""
        return os.getenv("NETRA_OTLP_ENDPOINT") or os.getenv("OTEL_EXPORTER_OTLP_ENDPOINT")

    def _parse_headers(self, headers: Optional[str]) -> Dict[str, str] | Any:
        """Parse headers from parameter or environment variable."""
        headers = headers or os.getenv("NETRA_HEADERS")
        if isinstance(headers, str):
            return parse_env_headers(headers)
        return {}

    def _validate_api_key(self) -> None:
        """Validate that API key exists for Netra endpoints."""
        if self.otlp_endpoint and "getnetra" in self.otlp_endpoint.lower() and not self.api_key:
            print("Error: Missing Netra API key, go to netra dashboard to create one")
            print("Set the NETRA_API_KEY environment variable to the key")

    def _setup_authentication(self) -> None:
        """Setup authentication headers based on endpoint and API key."""
        if not self.api_key or not self.otlp_endpoint:
            return

        is_netra = "getnetra" in self.otlp_endpoint.lower()
        auth_key = "x-api-key" if is_netra else "Authorization"
        auth_value = self.api_key if is_netra else f"Bearer {self.api_key}"

        if not self.headers:
            self.headers = {auth_key: auth_value}
        elif auth_key not in self.headers:
            self.headers[auth_key] = auth_value

    def _get_bool_config(self, param: Optional[bool], env_var: str, default: bool) -> bool:
        """Get boolean configuration from parameter or environment variable."""
        if param is not None:
            return param

        env_value = os.getenv(env_var)
        if env_value is None:
            return default

        return env_value.lower() in ("1", "true")

    def _get_resource_attributes(self, resource_attributes: Optional[Dict[str, Any]]) -> Dict[str, Any] | Any:
        """Get resource attributes from parameter or environment variable."""
        if resource_attributes is not None:
            return resource_attributes

        env_ra = os.getenv("NETRA_RESOURCE_ATTRS")
        if not env_ra:
            return {}

        try:
            return json.loads(env_ra)
        except (json.JSONDecodeError, ValueError) as e:
            import logging

            logger = logging.getLogger(__name__)
            logger.warning(f"Failed to parse NETRA_RESOURCE_ATTRS: {e}")
            return {}

    def _set_trace_content_env(self) -> None:
        """Set TRACELOOP_TRACE_CONTENT environment variable based on trace_content."""
        os.environ["TRACELOOP_TRACE_CONTENT"] = "true" if self.trace_content else "false"



# File: netra/exceptions/injection.py

from typing import Dict, List, Optional


class InjectionException(Exception):
    """
    Raised when prompt injection is detected in input and blocking is enabled.

    Attributes:
        message (str): Human-readable explanation of why blocking occurred.
        has_violation (bool): True if prompt injection was detected in the provided text.
        violations (List[str]): List of violation types that were detected.
        is_blocked (bool): True if blocking is enabled and prompt injection was detected.
        violation_actions (Dict[str, List[str]]): Dictionary mapping action types to lists of violations.
    """

    def __init__(
        self,
        message: str = "Input blocked due to detected injection.",
        has_violation: bool = True,
        violations: Optional[List[str]] = None,
        is_blocked: bool = True,
        violation_actions: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        """
        Initialize the injection exception.

        Args:
            message: The message to display.
            has_violation: Whether a violation was detected.
            violations: List of violations detected.
            is_blocked: Whether the input was blocked.
            violation_actions: Dictionary mapping action types to lists of violations.
        """
        # Always pass the message to the base Exception constructor
        super().__init__(message)

        # Store structured attributes
        self.has_violation: bool = has_violation
        self.violations: List[str] = violations or []
        self.is_blocked: bool = is_blocked
        self.violation_actions: Dict[str, List[str]] = violation_actions or {}



from __future__ import annotations

import logging

_LOG_FORMATTER = logging.Formatter(
    fmt="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


def configure_package_logging(debug_mode: bool) -> None:
    """
    Configure logging for the netra package and related dependencies.

    Args:
        debug_mode: Whether to enable debug logging.
    """

    pkg_logger = logging.getLogger("netra")
    httpx_logger = logging.getLogger("httpx")
    httpcore_logger = logging.getLogger("httpcore")
    otel_trace_logger = logging.getLogger("opentelemetry.trace")
    otel_instr_logger = logging.getLogger("opentelemetry.instrumentation")
    otel_instr_instrumentor_logger = logging.getLogger("opentelemetry.instrumentation.instrumentor")

    # Always clear Netra handlers so we can reconfigure cleanly
    pkg_logger.handlers.clear()

    if debug_mode:
        # Netra: verbose
        pkg_logger.setLevel(logging.DEBUG)
        handler = logging.StreamHandler()
        handler.setFormatter(_LOG_FORMATTER)
        pkg_logger.addHandler(handler)
        pkg_logger.propagate = False

        # HTTP clients: visible but not too noisy
        httpx_logger.setLevel(logging.INFO)
        httpcore_logger.setLevel(logging.INFO)
        httpx_logger.propagate = True
        httpcore_logger.propagate = True

        # OpenTelemetry: allow warnings and above
        otel_trace_logger.setLevel(logging.WARNING)
        otel_instr_logger.setLevel(logging.WARNING)
        otel_instr_instrumentor_logger.setLevel(logging.WARNING)
        otel_trace_logger.propagate = True
        otel_instr_logger.propagate = True
        otel_instr_instrumentor_logger.propagate = True
    else:
        # Netra: completely silent unless user opts in
        pkg_logger.setLevel(logging.CRITICAL)
        pkg_logger.addHandler(logging.NullHandler())
        pkg_logger.propagate = False

        # HTTP clients: keep their own warnings local, do not bubble via Netra
        httpx_logger.setLevel(logging.WARNING)
        httpcore_logger.setLevel(logging.WARNING)
        httpx_logger.propagate = False
        httpcore_logger.propagate = False

        # OpenTelemetry: suppress logs triggered by Netra usage
        otel_trace_logger.setLevel(logging.ERROR)
        otel_instr_logger.setLevel(logging.ERROR)
        otel_instr_instrumentor_logger.setLevel(logging.ERROR)
        otel_trace_logger.propagate = False
        otel_instr_logger.propagate = False
        otel_instr_instrumentor_logger.propagate = False



import logging
import time
from collections.abc import Awaitable
from typing import Any, AsyncIterator, Callable, Dict, Iterator, Tuple

from opentelemetry import context as context_api
from opentelemetry.trace import Span, SpanKind, Tracer, set_span_in_context
from opentelemetry.trace.status import Status, StatusCode
from wrapt import ObjectProxy

from netra.instrumentation.openai.utils import (
    model_as_dict,
    set_request_attributes,
    set_response_attributes,
    should_suppress_instrumentation,
)

logger = logging.getLogger(__name__)

# Span names
CHAT_SPAN_NAME = "openai.chat"
EMBEDDING_SPAN_NAME = "openai.embedding"
RESPONSE_SPAN_NAME = "openai.response"


def chat_wrapper(tracer: Tracer) -> Callable[..., Any]:
    """Wrapper for chat completions"""

    def wrapper(wrapped: Callable[..., Any], instance: Any, args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> Any:
        if should_suppress_instrumentation():
            return wrapped(*args, **kwargs)

        is_streaming = kwargs.get("stream", False)
        if is_streaming:
            span = tracer.start_span(CHAT_SPAN_NAME, kind=SpanKind.CLIENT, attributes={"llm.request.type": "chat"})
            try:
                context = context_api.attach(set_span_in_context(span))
                set_request_attributes(span, kwargs, "chat")
                start_time = time.time()
                response = wrapped(*args, **kwargs)
                return StreamingWrapper(span=span, response=response, start_time=start_time, request_kwargs=kwargs)
            except Exception as e:
                logger.error("netra.instrumentation.openai: %s", e)
                span.set_status(Status(StatusCode.ERROR, str(e)))
                span.record_exception(e)
                span.end()
                raise
            finally:
                context_api.detach(context)

        else:
            with tracer.start_as_current_span(
                CHAT_SPAN_NAME, kind=SpanKind.CLIENT, attributes={"llm.request.type": "chat"}
            ) as span:
                try:
                    set_request_attributes(span, kwargs, "chat")
                    start_time = time.time()
                    response = wrapped(*args, **kwargs)
                    end_time = time.time()
                    response_dict = model_as_dict(response)
                    set_response_attributes(span, response_dict)
                    duration = end_time - start_time
                    span.set_attribute("llm.response.duration", duration)
                    span.set_attribute("gen_ai.performance.time_to_first_token", duration)
                    span.set_status(Status(StatusCode.OK))
                    return response
                except Exception as e:
                    logger.error("netra.instrumentation.openai: %s", e)
                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    raise

    return wrapper


def achat_wrapper(tracer: Tracer) -> Callable[..., Awaitable[Any]]:
    """Async wrapper for chat completions"""

    async def wrapper(
        wrapped: Callable[..., Awaitable[Any]], instance: Any, args: Tuple[Any, ...], kwargs: Dict[str, Any]
    ) -> Any:
        if should_suppress_instrumentation():
            return await wrapped(*args, **kwargs)

        is_streaming = kwargs.get("stream", False)
        if is_streaming:
            span = tracer.start_span(CHAT_SPAN_NAME, kind=SpanKind.CLIENT, attributes={"llm.request.type": "chat"})
            try:
                context = context_api.attach(set_span_in_context(span))
                set_request_attributes(span, kwargs, "chat")
                start_time = time.time()
                response = await wrapped(*args, **kwargs)
                return AsyncStreamingWrapper(span=span, response=response, start_time=start_time, request_kwargs=kwargs)
            except Exception as e:
                logger.error("netra.instrumentation.openai: %s", e)
                span.set_status(Status(StatusCode.ERROR, str(e)))
                span.record_exception(e)
                span.end()
                raise
            finally:
                context_api.detach(context)
        else:
            with tracer.start_as_current_span(
                CHAT_SPAN_NAME, kind=SpanKind.CLIENT, attributes={"llm.request.type": "chat"}
            ) as span:
                try:
                    set_request_attributes(span, kwargs, "chat")
                    start_time = time.time()
                    response = await wrapped(*args, **kwargs)
                    end_time = time.time()
                    response_dict = model_as_dict(response)
                    set_response_attributes(span, response_dict)
                    duration = end_time - start_time
                    span.set_attribute("llm.response.duration", duration)
                    span.set_attribute("gen_ai.performance.time_to_first_token", duration)
                    span.set_status(Status(StatusCode.OK))
                    return response
                except Exception as e:
                    logger.error("netra.instrumentation.openai: %s", e)
                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    raise

    return wrapper


def embeddings_wrapper(tracer: Tracer) -> Callable[..., Any]:
    """Wrapper for embeddings"""

    def wrapper(wrapped: Callable[..., Any], instance: Any, args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> Any:
        if should_suppress_instrumentation():
            return wrapped(*args, **kwargs)

        with tracer.start_as_current_span(
            EMBEDDING_SPAN_NAME, kind=SpanKind.CLIENT, attributes={"llm.request.type": "embedding"}
        ) as span:
            try:
                set_request_attributes(span, kwargs, "embedding")
                start_time = time.time()
                response = wrapped(*args, **kwargs)
                end_time = time.time()
                response_dict = model_as_dict(response)
                set_response_attributes(span, response_dict)
                duration = end_time - start_time
                span.set_attribute("llm.response.duration", duration)
                span.set_attribute("gen_ai.performance.time_to_first_token", duration)
                span.set_status(Status(StatusCode.OK))
                return response
            except Exception as e:
                logger.error("netra.instrumentation.openai: %s", e)
                span.set_status(Status(StatusCode.ERROR, str(e)))
                raise

    return wrapper


def aembeddings_wrapper(tracer: Tracer) -> Callable[..., Awaitable[Any]]:
    """Async wrapper for embeddings"""

    async def wrapper(
        wrapped: Callable[..., Awaitable[Any]], instance: Any, args: Tuple[Any, ...], kwargs: Dict[str, Any]
    ) -> Any:
        if should_suppress_instrumentation():
            return await wrapped(*args, **kwargs)

        with tracer.start_as_current_span(
            EMBEDDING_SPAN_NAME, kind=SpanKind.CLIENT, attributes={"llm.request.type": "embedding"}
        ) as span:
            try:
                set_request_attributes(span, kwargs, "embedding")
                start_time = time.time()
                response = await wrapped(*args, **kwargs)
                end_time = time.time()
                response_dict = model_as_dict(response)
                set_response_attributes(span, response_dict)
                duration = end_time - start_time
                span.set_attribute("llm.response.duration", duration)
                span.set_attribute("gen_ai.performance.time_to_first_token", duration)
                span.set_status(Status(StatusCode.OK))
                return response
            except Exception as e:
                logger.error("netra.instrumentation.openai: %s", e)
                span.set_status(Status(StatusCode.ERROR, str(e)))
                raise

    return wrapper


def responses_wrapper(tracer: Tracer) -> Callable[..., Any]:
    """Wrapper for responses.create (new OpenAI API)"""

    def wrapper(wrapped: Callable[..., Any], instance: Any, args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> Any:
        if should_suppress_instrumentation():
            return wrapped(*args, **kwargs)

        is_streaming = kwargs.get("stream", False)
        if is_streaming:
            span = tracer.start_span(
                RESPONSE_SPAN_NAME, kind=SpanKind.CLIENT, attributes={"llm.request.type": "response"}
            )
            try:
                context = context_api.attach(set_span_in_context(span))
                set_request_attributes(span, kwargs, "response")
                start_time = time.time()
                response = wrapped(*args, **kwargs)
                return StreamingWrapper(span=span, response=response, start_time=start_time, request_kwargs=kwargs)
            except Exception as e:
                logger.error("netra.instrumentation.openai: %s", e)
                span.set_status(Status(StatusCode.ERROR, str(e)))
                span.record_exception(e)
                span.end()
                raise
            finally:
                context_api.detach(context)
        else:
            with tracer.start_as_current_span(
                RESPONSE_SPAN_NAME, kind=SpanKind.CLIENT, attributes={"llm.request.type": "response"}
            ) as span:
                try:
                    set_request_attributes(span, kwargs, "response")
                    start_time = time.time()
                    response = wrapped(*args, **kwargs)
                    end_time = time.time()
                    response_dict = model_as_dict(response)
                    set_response_attributes(span, response_dict)
                    duration = end_time - start_time
                    span.set_attribute("llm.response.duration", duration)
                    span.set_attribute("gen_ai.performance.time_to_first_token", duration)
                    span.set_status(Status(StatusCode.OK))
                    return response
                except Exception as e:
                    logger.error("netra.instrumentation.openai: %s", e)
                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    raise

    return wrapper


def aresponses_wrapper(tracer: Tracer) -> Callable[..., Awaitable[Any]]:
    """Async wrapper for responses.create (new OpenAI API)"""

    async def wrapper(wrapped: Callable[..., Awaitable[Any]], instance: Any, args: Any, kwargs: Dict[str, Any]) -> Any:
        if should_suppress_instrumentation():
            return await wrapped(*args, **kwargs)

        is_streaming = kwargs.get("stream", False)
        if is_streaming:
            span = tracer.start_span(
                RESPONSE_SPAN_NAME, kind=SpanKind.CLIENT, attributes={"llm.request.type": "response"}
            )
            try:
                context = context_api.attach(set_span_in_context(span))
                set_request_attributes(span, kwargs, "response")
                start_time = time.time()
                response = await wrapped(*args, **kwargs)
                return AsyncStreamingWrapper(span=span, response=response, start_time=start_time, request_kwargs=kwargs)
            except Exception as e:
                logger.error("netra.instrumentation.openai: %s", e)
                span.set_status(Status(StatusCode.ERROR, str(e)))
                span.record_exception(e)
                span.end()
                raise
            finally:
                context_api.detach(context)
        else:
            with tracer.start_as_current_span(
                RESPONSE_SPAN_NAME, kind=SpanKind.CLIENT, attributes={"llm.request.type": "response"}
            ) as span:
                try:
                    set_request_attributes(span, kwargs, "response")
                    start_time = time.time()
                    response = await wrapped(*args, **kwargs)
                    end_time = time.time()
                    response_dict = model_as_dict(response)
                    set_response_attributes(span, response_dict)
                    duration = end_time - start_time
                    span.set_attribute("llm.response.duration", duration)
                    span.set_attribute("gen_ai.performance.time_to_first_token", duration)
                    span.set_status(Status(StatusCode.OK))
                    return response
                except Exception as e:
                    logger.error("netra.instrumentation.openai: %s", e)
                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    raise

    return wrapper


class StreamingWrapper(ObjectProxy):  # type: ignore[misc]
    """Wrapper for streaming responses"""

    def __init__(self, span: Span, response: Iterator[Any], start_time: float, request_kwargs: Dict[str, Any]) -> None:
        super().__init__(response)
        self._span = span
        self._start_time = start_time
        self._request_kwargs = request_kwargs
        self._complete_response: Dict[str, Any] = {"choices": [], "model": ""}
        self._first_content_recorded: bool = False

    def _is_chat(self) -> bool:
        """Determine if the request is a chat request."""
        return isinstance(self._request_kwargs, dict) and "messages" in self._request_kwargs

    def _ensure_choice(self, index: int) -> None:
        """Ensure choices list has an entry at index."""
        while len(self._complete_response["choices"]) <= index:
            if self._is_chat():
                self._complete_response["choices"].append({"message": {"role": "assistant", "content": ""}})
            else:
                self._complete_response["choices"].append({"text": ""})

    def __enter__(self) -> "StreamingWrapper":
        if hasattr(self.__wrapped__, "__enter__"):
            self.__wrapped__.__enter__()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if hasattr(self.__wrapped__, "__exit__"):
            self.__wrapped__.__exit__(exc_type, exc_val, exc_tb)
        if exc_type is not None:
            self._span.set_status(Status(StatusCode.ERROR, str(exc_val)))
            self._span.record_exception(exc_val)
            self._span.end()

    def __iter__(self) -> Iterator[Any]:
        return self

    def __next__(self) -> Any:
        try:
            chunk = self.__wrapped__.__next__()
            self._process_chunk(chunk)
            return chunk
        except StopIteration:
            self._finalize_span()
            raise

    def _process_chunk(self, chunk: Any) -> None:
        """Process streaming chunk"""
        chunk_dict = model_as_dict(chunk)

        if chunk_dict.get("model"):
            self._complete_response["model"] = chunk_dict["model"]

        choices = chunk_dict.get("choices") or []

        # Completion API
        if isinstance(choices, list):
            for choice in choices:
                index = int(choice.get("index", 0))
                self._ensure_choice(index)
                delta = choice.get("delta") or {}
                content_piece = None
                if isinstance(delta, dict) and delta.get("content"):
                    content_piece = str(delta.get("content", ""))
                    if content_piece and not self._first_content_recorded:
                        self._first_content_recorded = True
                        self._span.set_attribute(
                            "gen_ai.performance.time_to_first_token", time.time() - self._start_time
                        )
                    self._complete_response["choices"][index].setdefault(
                        "message", {"role": "assistant", "content": ""}
                    )
                    self._complete_response["choices"][index]["message"]["content"] += content_piece

                if choice.get("finish_reason"):
                    self._complete_response["choices"][index]["finish_reason"] = choice.get("finish_reason")

        if chunk_dict.get("usage") and isinstance(chunk_dict["usage"], dict):
            self._complete_response["usage"] = chunk_dict["usage"]

        # Response API
        if chunk_dict.get("delta") and not self._first_content_recorded:
            self._first_content_recorded = True
            self._span.set_attribute("gen_ai.performance.time_to_first_token", time.time() - self._start_time)
        if chunk_dict.get("response"):
            response = chunk_dict.get("response", {})
            if response.get("status") == "completed":
                response_output = response.get("output", {})
                for output in response_output:
                    content = output.get("content")
                    for index, chunk in enumerate(content):
                        assistant_text = chunk.get("text", "")
                        self._complete_response["choices"] = [
                            {"message": {"role": "assistant", "content": assistant_text}}
                        ]

                usage = response.get("usage", {})
                self._complete_response["usage"] = usage

        self._span.add_event("llm.content.completion.chunk")

    def _finalize_span(self) -> None:
        """Finalize span when streaming is complete"""
        end_time = time.time()
        duration = end_time - self._start_time
        set_response_attributes(self._span, self._complete_response)
        self._span.set_attribute("llm.response.duration", duration)
        self._span.set_status(Status(StatusCode.OK))
        self._span.end()


class AsyncStreamingWrapper(ObjectProxy):  # type: ignore[misc]
    """Async wrapper for streaming responses"""

    def __init__(
        self, span: Span, response: AsyncIterator[Any], start_time: float, request_kwargs: Dict[str, Any]
    ) -> None:
        super().__init__(response)
        self._span = span
        self._start_time = start_time
        self._request_kwargs = request_kwargs
        self._complete_response: Dict[str, Any] = {"choices": [], "model": ""}
        self._first_content_recorded: bool = False

    def _is_chat(self) -> bool:
        """Determine if the request is a chat request."""
        return isinstance(self._request_kwargs, dict) and "messages" in self._request_kwargs

    def _ensure_choice(self, index: int) -> None:
        """Ensure choices list has an entry at index."""
        while len(self._complete_response["choices"]) <= index:
            if self._is_chat():
                self._complete_response["choices"].append({"message": {"role": "assistant", "content": ""}})
            else:
                self._complete_response["choices"].append({"text": ""})

    async def __aenter__(self) -> "AsyncStreamingWrapper":
        if hasattr(self.__wrapped__, "__aenter__"):
            await self.__wrapped__.__aenter__()
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if hasattr(self.__wrapped__, "__aexit__"):
            await self.__wrapped__.__aexit__(exc_type, exc_val, exc_tb)
        if exc_type is not None:
            self._span.set_status(Status(StatusCode.ERROR, str(exc_val)))
            self._span.record_exception(exc_val)
            self._span.end()

    def __aiter__(self) -> AsyncIterator[Any]:
        return self

    async def __anext__(self) -> Any:
        try:
            chunk = await self.__wrapped__.__anext__()
            self._process_chunk(chunk)
            return chunk
        except StopAsyncIteration:
            self._finalize_span()
            raise

    def _process_chunk(self, chunk: Any) -> None:
        """Process streaming chunk"""
        chunk_dict = model_as_dict(chunk)

        if chunk_dict.get("model"):
            self._complete_response["model"] = chunk_dict["model"]

        choices = chunk_dict.get("choices") or []

        # Completion API
        if isinstance(choices, list):
            for choice in choices:
                index = int(choice.get("index", 0))
                self._ensure_choice(index)
                delta = choice.get("delta") or {}
                content_piece = None
                if isinstance(delta, dict) and delta.get("content"):
                    content_piece = str(delta.get("content", ""))
                    if content_piece and not self._first_content_recorded:
                        self._first_content_recorded = True
                        self._span.set_attribute(
                            "gen_ai.performance.time_to_first_token", time.time() - self._start_time
                        )
                    self._complete_response["choices"][index].setdefault(
                        "message", {"role": "assistant", "content": ""}
                    )
                    self._complete_response["choices"][index]["message"]["content"] += content_piece

                if choice.get("finish_reason"):
                    self._complete_response["choices"][index]["finish_reason"] = choice.get("finish_reason")

        if chunk_dict.get("usage") and isinstance(chunk_dict["usage"], dict):
            self._complete_response["usage"] = chunk_dict["usage"]

        # Response API
        if chunk_dict.get("delta") and not self._first_content_recorded:
            self._first_content_recorded = True
            self._span.set_attribute("gen_ai.performance.time_to_first_token", time.time() - self._start_time)
        if chunk_dict.get("response"):
            response = chunk_dict.get("response", {})
            if response.get("status") == "completed":
                response_output = response.get("output", {})
                for output in response_output:
                    content = output.get("content")
                    for index, chunk in enumerate(content):
                        assistant_text = chunk.get("text", "")
                        self._complete_response["choices"] = [
                            {"message": {"role": "assistant", "content": assistant_text}}
                        ]

                usage = response.get("usage", {})
                self._complete_response["usage"] = usage

        self._span.add_event("llm.content.completion.chunk")

    def _finalize_span(self) -> None:
        """Finalize span when streaming is complete"""
        end_time = time.time()
        duration = end_time - self._start_time
        set_response_attributes(self._span, self._complete_response)
        self._span.set_attribute("llm.response.duration", duration)
        self._span.set_status(Status(StatusCode.OK))
        self._span.end()



from netra.exporters.filtering_span_exporter import FilteringSpanExporter
from netra.exporters.trial_aware_otlp_exporter import TrialAwareOTLPExporter

__all__ = ["FilteringSpanExporter", "TrialAwareOTLPExporter"]



import wrapt
import logging
from opentelemetry.trace import Tracer, get_tracer
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor

from netra.instrumentation.claude_agent_sdk.version import __version__
from netra.instrumentation.claude_agent_sdk.wrappers import (
    client_query_wrapper, 
    client_response_wrapper, 
    query_wrapper
)

logger = logging.getLogger(__name__)

_instruments = ("claude_agent_sdk >= 0.1.0", )

class NetraClaudeAgentSDKInstrumentor(BaseInstrumentor):
    def instrumentation_dependencies(self):
        """
        Return the list of packages required for this instrumentation to function.

        Args:
            None

        Returns:
            tuple: A tuple of pip requirement strings for the instrumented library.
        """
        return _instruments
    
    def _instrument(self, **kwargs):
        """
        Set up OpenTelemetry instrumentation for the Claude Agent SDK.

        Wraps InternalClient.process_query, ClaudeSDKClient.query, and
        ClaudeSDKClient.receive_messages with tracing wrappers.

        Args:
            **kwargs: Accepts an optional 'tracer_provider' (TracerProvider) to use
                      instead of the global provider.

        Returns:
            None
        """
        try:
            tracer_provider = kwargs.get("tracer_provider")
            tracer = get_tracer(__name__, __version__, tracer_provider)
        except Exception as e:
            logger.error(f"Failed to initialize tracer: {e}")
            return
        self._instrument_query(tracer)
        self._instrument_client_query(tracer)
        self._instrument_client_response(tracer)

    def _uninstrument(self, **kwargs):
        """
        Remove all custom instrumentation wrappers from the Claude Agent SDK.

        Args:
            **kwargs: Not used; accepted for compatibility with BaseInstrumentor interface.

        Returns:
            None
        """
        self._uninstrument_query()
        self._uninstrument_client_query()
        self._uninstrument_client_response()

    def _instrument_query(self, tracer: Tracer):
        """
        Wrap InternalClient.process_query with a tracing wrapper.

        Args:
            tracer (Tracer): The OpenTelemetry tracer to pass to the query wrapper.

        Returns:
            None
        """
        try:
            wrapt.wrap_function_wrapper(
                "claude_agent_sdk._internal.client", 
                "InternalClient.process_query", 
                query_wrapper(tracer) 
            )
        except Exception as e:
            logger.error(f"Failed to instrument claude-agent-sdk query: {e}")

    def _instrument_client_query(self, tracer: Tracer):
        """
        Wrap ClaudeSDKClient.query to capture the prompt for downstream tracing.

        Args:
            tracer (Tracer): The OpenTelemetry tracer (accepted for interface consistency).

        Returns:
            None
        """
        try:
            wrapt.wrap_function_wrapper(
                "claude_agent_sdk.client", 
                "ClaudeSDKClient.query", 
                client_query_wrapper() 
            )
        except Exception as e:
            logger.error(f"Failed to instrument claude-sdk-client query: {e}")

    def _instrument_client_response(self, tracer: Tracer):
        """
        Wrap ClaudeSDKClient.receive_messages with a tracing wrapper.

        Args:
            tracer (Tracer): The OpenTelemetry tracer to pass to the response wrapper.

        Returns:
            None
        """
        try:
            wrapt.wrap_function_wrapper(
                "claude_agent_sdk.client", 
                "ClaudeSDKClient.receive_messages", 
                client_response_wrapper(tracer) 
            )
        except Exception as e:
            logger.error(f"Failed to instrument claude-sdk-client response: {e}")

    def _uninstrument_query(self):
        """
        Remove the tracing wrapper from InternalClient.process_query.

        Args:
            None

        Returns:
            None
        """
        try:
            unwrap("claude_agent_sdk._internal.client", "InternalClient.process_query")
        except (AttributeError, ModuleNotFoundError):
            logger.error(f"Failed to uninstrument claude-agent-sdk query")

    def _uninstrument_client_query(self):
        """
        Remove the tracing wrapper from ClaudeSDKClient.query.

        Args:
            None

        Returns:
            None
        """
        try:
            unwrap("claude_agent_sdk.client", "ClaudeSDKClient.query")
        except (AttributeError, ModuleNotFoundError):
            logger.error(f"Failed to uninstrument claude-sdk-client query")

    def _uninstrument_client_response(self):
        """
        Remove the tracing wrapper from ClaudeSDKClient.receive_messages.

        Args:
            None

        Returns:
            None
        """
        try:
            unwrap("claude_agent_sdk.client", "ClaudeSDKClient.receive_messages")
        except (AttributeError, ModuleNotFoundError):
            logger.error(f"Failed to uninstrument claude-sdk-client response")


import logging
from typing import Any, Collection

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.trace import get_tracer
from wrapt import wrap_function_wrapper

from netra.instrumentation.cerebras.version import __version__
from netra.instrumentation.cerebras.wrappers import (
    achat_wrapper,
    acompletions_wrapper,
    chat_wrapper,
    completions_wrapper,
)

logger = logging.getLogger(__name__)

_instruments = ("cerebras-cloud-sdk >= 1.0.0",)


class NetraCerebrasInstrumentor(BaseInstrumentor):  # type:ignore[misc]
    """Custom Cerebras instrumentor for Netra SDK."""

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs: Any) -> None:
        tracer = get_tracer(__name__, __version__, kwargs.get("tracer_provider"))

        try:
            wrap_function_wrapper(
                "cerebras.cloud.sdk.resources.chat.completions",
                "CompletionsResource.create",
                chat_wrapper(tracer),
            )
            wrap_function_wrapper(
                "cerebras.cloud.sdk.resources.chat.completions",
                "AsyncCompletionsResource.create",
                achat_wrapper(tracer),
            )
        except Exception as e:
            logger.error(f"Failed to instrument Cerebras chat completions: {e}")

        try:
            wrap_function_wrapper(
                "cerebras.cloud.sdk.resources.completions",
                "CompletionsResource.create",
                completions_wrapper(tracer),
            )
            wrap_function_wrapper(
                "cerebras.cloud.sdk.resources.completions",
                "AsyncCompletionsResource.create",
                acompletions_wrapper(tracer),
            )
        except Exception as e:
            logger.error(f"Failed to instrument Cerebras completions: {e}")

    def _uninstrument(self, **kwargs: Any) -> None:
        try:
            unwrap("cerebras.cloud.sdk.resources.chat.completions", "CompletionsResource.create")
            unwrap("cerebras.cloud.sdk.resources.chat.completions", "AsyncCompletionsResource.create")
        except (AttributeError, ModuleNotFoundError):
            logger.error("Failed to uninstrument Cerebras chat completions")

        try:
            unwrap("cerebras.cloud.sdk.resources.completions", "CompletionsResource.create")
            unwrap("cerebras.cloud.sdk.resources.completions", "AsyncCompletionsResource.create")
        except (AttributeError, ModuleNotFoundError):
            logger.error("Failed to uninstrument Cerebras completions")



import logging
from typing import Any, AsyncIterator, Callable, Tuple
from opentelemetry import trace
from opentelemetry.context import Context
from opentelemetry.trace import Span, SpanKind, Tracer
from opentelemetry.trace.status import Status, StatusCode
from claude_agent_sdk import SystemMessage, AssistantMessage, UserMessage, ResultMessage

from netra.instrumentation.claude_agent_sdk.utils import (
    set_request_attributes,
    set_system_message_attributes,
    set_assistant_message_attributes,
    set_user_message_attributes,
    set_result_message_attributes,
)

logger = logging.getLogger(__name__)

QUERY_SPAN_NAME = "claude-agent.query"
AGENT_CONVERSATION_SPAN_NAME = "claude-agent.conversation"


async def _dispatch_messages(
    tracer: Tracer,
    root_span: Span,
    root_ctx: Context,
    aiterator: AsyncIterator,
    prompt_index: int,
) -> AsyncIterator:
    """
    Dispatch each incoming SDK message to its span attribute handler and yield it.

    Args:
        tracer (Tracer): The OpenTelemetry tracer for creating child spans.
        root_span (Span): The root span to attach message attributes to.
        root_ctx (Context): The root span context for child span parenting.
        aiterator (AsyncIterator): The async iterator of SDK messages to process.
        prompt_index (int): The current prompt index passed to the result message attribute setter.

    Returns:
        AsyncIterator: Yields each SDK message after processing its span attributes.
    """
    while True:
        try:
            message = await anext(aiterator)
        except StopAsyncIteration:
            return

        try:
            if isinstance(message, SystemMessage):
                set_system_message_attributes(root_span, message)
            elif isinstance(message, ResultMessage):
                set_result_message_attributes(root_span, message, prompt_index)
            elif isinstance(message, AssistantMessage):
                set_assistant_message_attributes(tracer, root_ctx, message)
            elif isinstance(message, UserMessage):
                set_user_message_attributes(tracer, root_ctx, message)
        except Exception as e:
            logger.error("Failed to record message attributes: %s", e)

        yield message


def query_wrapper(tracer: Tracer):
    """
    Return a wrapper that traces a single InternalClient.process_query call with child spans per message.

    Args:
        tracer (Tracer): The OpenTelemetry tracer to use for creating spans.

    Returns:
        Callable: An async generator wrapper function for InternalClient.process_query.
    """
    async def wrapper(
        wrapped: Callable[..., Any],
        instance: Any,
        args: Tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Any:
        """
        Wrap InternalClient.process_query to create a root span and dispatch per-message child spans.

        Args:
            wrapped (Callable): The original process_query function.
            instance (Any): The InternalClient instance.
            args (Tuple): Positional arguments
            kwargs (dict): Keyword arguments; may include 'prompt' and 'options'.

        Returns:
            AsyncIterator: Yields SDK messages with span instrumentation applied.
        """
        prompt = args[0] if len(args) > 0 else kwargs.get("prompt")
        options = args[1] if len(args) > 1 else kwargs.get("options")

        root_span = tracer.start_span(QUERY_SPAN_NAME, kind=SpanKind.CLIENT)
        root_ctx = trace.set_span_in_context(root_span)
        prompt_index = 0

        try:
            try:
                with trace.use_span(root_span, end_on_exit=False):
                    prompt_index = set_request_attributes(root_span, prompt, options)
            except Exception as e:
                logger.error("Instrumentation setup failed: %s", e)

            aiterator = aiter(wrapped(*args, **kwargs))
            async for message in _dispatch_messages(tracer, root_span, root_ctx, aiterator, prompt_index):
                yield message

        except GeneratorExit:
            raise
        except Exception as e:
            logger.error("netra.instrumentation.claude-agent-sdk: %s", e)
            try:
                root_span.record_exception(e)
                root_span.set_status(Status(StatusCode.ERROR, str(e)))
            except Exception as span_err:
                logger.error("Failed to record exception on span: %s", span_err)
            raise
        finally:
            try:
                root_span.end()
            except Exception as e:
                logger.error("Failed to end span: %s", e)

    return wrapper


def client_query_wrapper():
    """
    Return a wrapper that captures the prompt from ClaudeSDKClient.query for later tracing.

    Args:
        None

    Returns:
        Callable: An async wrapper function for ClaudeSDKClient.query.
    """
    async def wrapper(
        wrapped: Callable[..., Any],
        instance: Any,
        args: Tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Any:
        """
        Intercept ClaudeSDKClient.query to store the prompt on the instance for later tracing.

        Args:
            wrapped (Callable): The original query function.
            instance (Any): The ClaudeSDKClient instance; prompt is stored as _instrumentation_prompt.
            args (Tuple): Positional arguments
            kwargs (dict): Keyword arguments; may include 'prompt'.

        Returns:
            Any: The result of the original query function.
        """
        instance._instrumentation_prompt = args[0] if len(args) > 0 else kwargs.get("prompt")
        return await wrapped(*args, **kwargs)

    return wrapper


def client_response_wrapper(tracer: Tracer):
    """
    Return a wrapper that traces a full ClaudeSDKClient.receive_messages call covering all messages.

    Args:
        tracer (Tracer): The OpenTelemetry tracer to use for creating spans.

    Returns:
        Callable: An async generator wrapper function for ClaudeSDKClient.receive_messages.
    """
    async def wrapper(
        wrapped: Callable[..., Any],
        instance: Any,
        args: Tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Any:
        """
        Wrap ClaudeSDKClient.receive_messages to create a root span and dispatch per-message child spans.

        Args:
            wrapped (Callable): The original receive_messages function.
            instance (Any): The ClaudeSDKClient instance; prompt and options are read from it.
            args (Tuple): Positional arguments forwarded to the original function.
            kwargs (dict): Keyword arguments forwarded to the original function.

        Returns:
            AsyncIterator: Yields SDK messages with span instrumentation applied.
        """
        prompt = getattr(instance, "_instrumentation_prompt", None)
        options = getattr(instance, "options", None)

        root_span = tracer.start_span(AGENT_CONVERSATION_SPAN_NAME, kind=SpanKind.CLIENT)
        root_ctx = trace.set_span_in_context(root_span)
        prompt_index = 0

        try:
            try:
                with trace.use_span(root_span, end_on_exit=False):
                    prompt_index = set_request_attributes(root_span, prompt, options)
            except Exception as e:
                logger.error("Instrumentation setup failed: %s", e)

            aiterator = aiter(wrapped(*args, **kwargs))
            async for message in _dispatch_messages(tracer, root_span, root_ctx, aiterator, prompt_index):
                yield message

        except GeneratorExit:
            raise
        except Exception as e:
            logger.error("netra.instrumentation.claude-agent-sdk: %s", e)
            try:
                root_span.record_exception(e)
                root_span.set_status(Status(StatusCode.ERROR, str(e)))
            except Exception as span_err:
                logger.error("Failed to record exception on span: %s", span_err)
            raise
        finally:
            try:
                root_span.end()
            except Exception as e:
                logger.error("Failed to end span: %s", e)

    return wrapper



from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel


class Scope(str, Enum):
    """Scope of data to query."""

    SPANS = "Spans"
    TRACES = "Traces"


class ChartType(str, Enum):
    """Type of chart visualization."""

    LINE_TIME_SERIES = "Line Time Series"
    BAR_TIME_SERIES = "Bar Time Series"
    HORIZONTAL_BAR = "Horizontal Bar"
    VERTICAL_BAR = "Vertical Bar"
    PIE = "Pie"
    NUMBER = "Number"


class Measure(str, Enum):
    """Metric to measure."""

    LATENCY = "Latency"
    ERROR_RATE = "Error Rate"
    PII_COUNT = "PII Count"
    REQUEST_COUNT = "Request Count"
    TOTAL_COST = "Total Cost"
    VIOLATIONS = "Violations"
    TOTAL_TOKENS = "Total Tokens"
    AUDIO_DURATION = "Audio Duration"
    CHARACTER_COUNT = "Character Count"
    TTFT = "Time to First Token"


class Aggregation(str, Enum):
    """Aggregation method for metrics."""

    AVERAGE = "Average"
    P50 = "p50"
    P90 = "p90"
    P95 = "p95"
    P99 = "p99"
    MEDIAN = "Median (p50)"
    PERCENTAGE = "Percentage"
    TOTAL_COUNT = "Total Count"


class GroupBy(str, Enum):
    """Time grouping granularity."""

    DAY = "day"
    HOUR = "hour"
    MINUTE = "minute"


class DimensionField(str, Enum):
    """Dimension fields for grouping results."""

    ENVIRONMENT = "environment"
    SERVICE = "service"
    MODEL_NAME = "model_name"


class Operator(str, Enum):
    """Filter operators for query conditions."""

    EQUALS = "equals"
    NOT_EQUALS = "not_equals"
    CONTAINS = "contains"
    NOT_CONTAINS = "not_contains"
    STARTS_WITH = "starts_with"
    ENDS_WITH = "ends_with"
    GREATER_THAN = "greater_than"
    LESS_THAN = "less_than"
    GREATER_EQUAL_TO = "greater_equal_to"
    LESS_EQUAL_TO = "less_equal_to"
    ANY_OF = "any_of"
    NONE_OF = "none_of"


class Type(str, Enum):
    """Data types for filter conditions."""

    STRING = "string"
    NUMBER = "number"
    BOOLEAN = "boolean"
    ARRAY_OPTIONS = "arrayOptions"
    OBJECT = "object"


class FilterField(str, Enum):
    """
    Filter fields for dashboard queries.

    Note:
        - Use MODEL_NAME for Spans scope
        - Use MODELS for Traces scope
        - For metadata filters, use metadata_field() helper function
    """

    TOTAL_COST = "total_cost"
    SERVICE = "service"
    TENANT_ID = "tenant_id"
    USER_ID = "user_id"
    SESSION_ID = "session_id"
    ENVIRONMENT = "environment"
    LATENCY = "latency"
    MODEL_NAME = "model_name"
    MODELS = "models"
    METADATA = "metadata"


class Filter(BaseModel):  # type:ignore[misc]
    """
    Filter condition for dashboard queries.

    Attributes:
        field: Filter field - use FilterField enum or metadata_field() helper.
        operator: Filter operator from FilterOperator enum.
        type: Data type from FilterType enum.
        value: The value to filter by.
        key: Required for FilterType.OBJECT filters.
    """

    field: FilterField
    operator: Operator
    type: Type
    value: Any
    key: Optional[str] = None


class Metrics(BaseModel):  # type:ignore[misc]
    """
    Metrics configuration for dashboard queries.

    Attributes:
        measure: The metric to measure (e.g., Metric.LATENCY).
        aggregation: The aggregation method (e.g., Aggregation.AVERAGE).
    """

    measure: Measure
    aggregation: Aggregation


class Dimension(BaseModel):  # type:ignore[misc]
    """
    Dimension configuration for dashboard queries.

    Attributes:
        field: The dimension field to group results by.
    """

    field: DimensionField


class FilterConfig(BaseModel):  # type:ignore[misc]
    """
    Filter configuration for dashboard queries.

    Attributes:
        start_time: Start time in ISO 8601 UTC format (YYYY-MM-DDTHH:mm:ss.SSSZ).
        end_time: End time in ISO 8601 UTC format (YYYY-MM-DDTHH:mm:ss.SSSZ).
        group_by: Time grouping granularity.
        filters: Optional list of filter conditions.
    """

    start_time: str
    end_time: str
    group_by: GroupBy
    filters: Optional[List[Filter]] = None


class TimeRange(BaseModel):  # type:ignore[misc]
    """Time range information in the response."""

    start_time: str
    end_time: str


class TimeSeriesDataPoint(BaseModel):  # type:ignore[misc]
    """Data point for time series without dimension."""

    date: str
    value: float


class Value(BaseModel):  # type:ignore[misc]
    """Value for a specific dimension."""

    dimension: str
    value: float


class TimeSeriesWithDimension(BaseModel):  # type:ignore[misc]
    """Time series data point with dimension values."""

    date: str
    values: List[Value]


class TimeSeriesResponse(BaseModel):  # type:ignore[misc]
    """Response for time series with dimension."""

    time_series: List[TimeSeriesWithDimension]
    dimensions: List[str]


class CategoricalDataPoint(BaseModel):  # type:ignore[misc]
    """Data point for categorical charts (Pie/Bar)."""

    dimension: str
    value: float


class NumberResponse(BaseModel):  # type:ignore[misc]
    """Response for number chart."""

    value: float


Data = Union[
    List[TimeSeriesDataPoint],
    TimeSeriesResponse,
    List[CategoricalDataPoint],
    NumberResponse,
    Dict[str, Any],
]


class QueryResponse(BaseModel):  # type:ignore[misc]
    """Response wrapper for dashboard queries."""

    time_range: TimeRange
    data: Data


class SessionFilterField(str, Enum):
    """Filter fields for session stats queries."""

    TENANT_ID = "tenant_id"
    ENVIRONMENT = "environment"
    SERVICE = "service"


class SessionFilterOperator(str, Enum):
    """Filter operators for session stats queries."""

    ANY_OF = "any_of"


class SessionFilterType(str, Enum):
    """Data types for session stats filter conditions."""

    ARRAY = "arrayOptions"


class SortField(str, Enum):
    """Sort fields for session stats queries."""

    SESSION_ID = "session_id"
    START_TIME = "start_time"
    TOTAL_REQUESTS = "totalRequests"
    TOTAL_COST = "totalCost"


class SortOrder(str, Enum):
    ASC = "asc"
    DESC = "desc"


class SessionFilter(BaseModel):  # type:ignore[misc]
    """
    Filter condition for session stats queries.

    Attributes:
        field: Filter field from SessionFilterField enum.
        operator: Filter operator from SessionFilterOperator enum.
        type: Data type from SessionFilterType enum.
        value: The list of values to filter by.
    """

    field: SessionFilterField
    operator: SessionFilterOperator
    type: SessionFilterType
    value: List[str]


class SessionStatsData(BaseModel):  # type:ignore[misc]
    """Individual session statistics data."""

    session_id: str
    start_time: str
    total_requests: int
    total_cost: float
    session_duration: str


class SessionStatsResult(BaseModel):  # type:ignore[misc]
    """Response wrapper for session stats queries."""

    data: List[Dict[str, Any]]
    has_next_page: bool
    next_cursor: Optional[str] = None


class SessionFilterConfig(BaseModel):  # type:ignore[misc]
    start_time: str
    end_time: str
    filters: Optional[List[SessionFilter]] = None



import logging
from typing import Any, Collection

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.trace import get_tracer
from wrapt import wrap_function_wrapper

from netra.instrumentation.cartesia.version import __version__
from netra.instrumentation.cartesia.wrappers import (
    stt_wrapper,
    stt_ws_wrapper,
    tts_bytes_wrapper,
    tts_ws_wrapper,
    voice_changer_wrapper,
)

logger = logging.getLogger(__name__)

_instruments = ("cartesia >= 2.0.17",)


class NetraCartesiaInstrumentor(BaseInstrumentor):  # type: ignore[misc]
    """
    Custom Cartesia instrumentor for Netra SDK:
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        """
        Return the instrument dependencies for this instrumentor.
        """
        return _instruments

    def _instrument(self, **kwargs: Any) -> None:
        """
        Instrument the Cartesia client methods.
        """
        try:
            tracer_provider = kwargs.get("tracer_provider")
            tracer = get_tracer(__name__, __version__, tracer_provider)
        except Exception as e:  # pragma: no cover - defensive
            logger.error(f"Failed to initialize Cartesia tracer: {e}")
            return

        try:
            wrap_function_wrapper(
                "cartesia.tts.client",
                "TtsClient.bytes",
                tts_bytes_wrapper(tracer),
            )
            wrap_function_wrapper(
                "cartesia.tts.client",
                "TtsClient.sse",
                tts_bytes_wrapper(tracer),
            )
            wrap_function_wrapper(
                "cartesia.tts.socket_client", "TtsClientWithWebsocket.websocket", tts_ws_wrapper(tracer)
            )
        except Exception as e:
            logger.error(f"Failed to instrument Cartesia tts utility: {e}")

        try:
            wrap_function_wrapper(
                "cartesia.stt.client",
                "SttClient.transcribe",
                stt_wrapper(tracer),
            )
            wrap_function_wrapper(
                "cartesia.stt.socket_client", "SttClientWithWebsocket.websocket", stt_ws_wrapper(tracer)
            )
        except Exception as e:
            logger.error(f"Failed to instrument Cartesia sst utility: {e}")

        try:
            wrap_function_wrapper(
                "cartesia.voice_changer.client",
                "VoiceChangerClient.bytes",
                voice_changer_wrapper(tracer),
            )
        except Exception as e:
            logger.error(f"Failed to instrument Cartesia voice changer utility: {e}")

    def _uninstrument(self, **kwargs: Any) -> None:
        """
        Uninstrument the Cartesia client methods.
        """
        try:
            unwrap("cartesia.tts.client", "TtsClient.bytes")
            unwrap("cartesia.tts.client", "TtsClient.sse")
            unwrap("cartesia.tts.client", "TtsClientWithWebsocket.websocket")
        except (AttributeError, ModuleNotFoundError):
            logger.error("Failed to uninstrument Cartesia tts utility")

        try:
            unwrap("cartesia.stt.client", "SttClient.transcribe")
            unwrap("cartesia.stt.client", "SttClientWithWebsocket.websocket")
        except (AttributeError, ModuleNotFoundError):
            logger.error("Failed to uninstrument Cartesia stt utility")

        try:
            unwrap("cartesia.voice_changer.client", "VoiceChangerClient.bytes")
        except (AttributeError, ModuleNotFoundError):
            logger.error("Failed to uninstrument Cartesia voice changer utility")



"""Data models for the simulation module."""

from dataclasses import dataclass
from enum import Enum
from typing import Optional


class ConversationStatus(Enum):
    """Status indicating whether to continue or stop the conversation."""

    CONTINUE = "continue"
    STOP = "stop"


@dataclass(slots=True, frozen=True)
class SimulationItem:
    """Represents a single item in a simulation run.

    Attributes:
        run_item_id: Unique identifier for the run item.
        message: The user message content.
        turn_id: Identifier for the conversation turn.
    """

    run_item_id: str
    message: str
    turn_id: str


@dataclass(slots=True)
class ConversationResponse:
    """Response from the conversation trigger API.

    Attributes:
        decision: The decision to continue or stop the conversation.
        reason: Optional reason for stopping the conversation.
        next_turn_id: Identifier for the next turn if continuing.
        next_user_message: The next user message if continuing.
        next_run_item_id: Identifier for the next run item if continuing.
    """

    decision: str
    reason: Optional[str] = None
    next_turn_id: Optional[str] = None
    next_user_message: Optional[str] = None
    next_run_item_id: Optional[str] = None


@dataclass(slots=True, frozen=True)
class TaskResult:
    """Result returned from the user's task function.

    Attributes:
        message: The response message from the task.
        session_id: The session identifier for conversation continuity.
    """

    message: str
    session_id: str



import asyncio
import logging
import os
import time
from typing import Any, Dict, List, Optional

import httpx

from netra.config import Config
from netra.evaluation.models import DatasetItem, TurnType

logger = logging.getLogger(__name__)


class EvaluationHttpClient:
    """
    Internal HTTP client for Evaluation APIs.
    """

    def __init__(self, config: Config) -> None:
        """
        Initialize HTTP client for evaluation endpoints.

        Args:
            config: The configuration object.
        """
        self._client: Optional[httpx.Client] = self._create_client(config)

    def _create_client(self, config: Config) -> Optional[httpx.Client]:
        """
        Create an HTTP client for evaluation endpoints.

        Args:
            config: The configuration object.

        Returns:
            An HTTP client for evaluation endpoints, or None if creation fails.
        """
        endpoint = (config.otlp_endpoint or "").strip()
        if not endpoint:
            logger.error("netra.evaluation: NETRA_OTLP_ENDPOINT is required for evaluation APIs")
            return None

        base_url = self._resolve_base_url(endpoint)
        headers = self._build_headers(config)
        timeout = self._get_timeout()

        try:
            return httpx.Client(base_url=base_url, headers=headers, timeout=timeout)
        except Exception as exc:
            logger.error("netra.evaluation: Failed to initialize evaluation HTTP client: %s", exc)
            return None

    def _resolve_base_url(self, endpoint: str) -> str:
        """
        Resolve base URL from endpoint.

        Args:
            endpoint: The endpoint to resolve.

        Returns:
            The resolved base URL.
        """
        base_url = endpoint.rstrip("/")
        if base_url.endswith("/telemetry"):
            base_url = base_url[: -len("/telemetry")]
        return base_url

    def _build_headers(self, config: Config) -> Dict[str, str]:
        """
        Build Headers for Evaluation Client

        Args:
            config: The configuration object.

        Returns:
            The headers for evaluation client.
        """
        headers: Dict[str, str] = dict(config.headers or {})
        api_key = config.api_key
        if api_key:
            headers["x-api-key"] = api_key
        return headers

    def _get_timeout(self) -> float:
        """
        Get timeout for evaluation client.

        Returns:
            The timeout for evaluation client.
        """
        timeout_env = os.getenv("NETRA_EVALUATION_TIMEOUT")
        if not timeout_env:
            return 10.0
        try:
            return float(timeout_env)
        except ValueError:
            logger.warning(
                "netra.evaluation: Invalid NETRA_EVALUATION_TIMEOUT value '%s', using default 10.0",
                timeout_env,
            )
            return 10.0

    def create_dataset(
        self, name: Optional[str], tags: Optional[List[str]] = None, turn_type: TurnType = TurnType.SINGLE
    ) -> Any:
        """
        Create an empty dataset

        Args:
            name: The name of the dataset.
            tags: Optional list of tags to associate with the dataset.
            turn_type: The turn type of the dataset, either "single" or "multi". Defaults to "single".

        Returns:
            A backend JSON response containing dataset info (id, name, tags, etc.) on success,
            or None if creation fails.
        """
        if not self._client:
            logger.error("netra.evaluation: Evaluation client is not initialized; cannot create dataset")
            return None
        try:
            url = "/evaluations/dataset"
            payload: Dict[str, Any] = {"name": name, "tags": tags if tags else [], "turnType": turn_type.value}
            response = self._client.post(url, json=payload)
            response.raise_for_status()
            data = response.json()
            if isinstance(data, dict) and "data" in data:
                logger.info("netra.evaluation: Dataset created successfully")
                return data.get("data", {})
        except Exception:
            response_json = response.json()
            logger.error(
                "netra.evaluation: Failed to create dataset: %s", response_json.get("error").get("message", "")
            )
            return None

    def add_dataset_item(self, dataset_id: str, item: DatasetItem) -> Any:
        """
        Add a single item to an existing dataset and return backend data (e.g., new item id).

        Args:
            dataset_id: The id of the dataset to which the item will be added.
            item_payload: The dataset item to add.

        Returns:
            A backend JSON response on success or {"success": False} on error.
        """
        if not self._client:
            logger.error("netra.evaluation: Evaluation client is not initialized; cannot add item to dataset")
            return {"success": False}
        try:
            url = f"/evaluations/dataset/{dataset_id}/items"
            item_payload: Dict[str, Any] = {
                "input": item.input if item.input else None,
                "expectedOutput": item.expected_output if item.expected_output else None,
                "tags": item.tags if item.tags else None,
                "metadata": item.metadata if item.metadata else None,
            }
            response = self._client.post(url, json=item_payload)
            response.raise_for_status()
            data = response.json()
            if isinstance(data, dict) and "data" in data:
                logger.info("netra.evaluation: Dataset item added successfully")
                return data.get("data", {})
        except Exception:
            response_json = response.json()
            logger.error(
                "netra.evaluation: Failed to add item to dataset '%s': %s",
                dataset_id,
                response_json.get("error").get("message", ""),
            )
            return None

    def get_dataset(self, dataset_id: str) -> Any:
        """
        Fetch dataset items for a dataset id.

        Args:
            dataset_id: The id of the dataset to fetch.

        Returns:
            A list of dataset items.
        """
        if not self._client:
            logger.error("netra.evaluation: Evaluation client is not initialized; cannot fetch dataset")
            return {"success": False}
        try:
            url = f"/evaluations/dataset/{dataset_id}"
            response = self._client.get(url)
            response.raise_for_status()
            data = response.json()
            if isinstance(data, dict) and "data" in data:
                logger.info("netra.evaluation: Dataset fetched successfully")
                return data.get("data", [])
        except Exception:
            response_json = response.json()
            logger.error(
                "netra.evaluation: Failed to fetch dataset '%s': %s",
                dataset_id,
                response_json.get("error").get("message", ""),
            )
            return None

    def create_run(
        self,
        name: str,
        dataset_id: Optional[str] = None,
        evaluators_config: Optional[List[Dict[str, Any]]] = None,
    ) -> Any:
        """
        Create a new run based on the provided name, dataset_id, and evaluators_config.

        Args:
            name: The name of the run.
            dataset_id: The id of the dataset to which the run will be associated.
            evaluators_config: Optional list of evaluators to be used for the run.

        Returns:
            A backend JSON response containing run_id
        """
        if not self._client:
            logger.error("netra.evaluation: Evaluation client is not initialized; cannot create run")
            return {"success": False}
        try:
            url = f"/evaluations/test_run"
            payload: Dict[str, Any] = {
                "name": name,
                "datasetId": dataset_id if dataset_id else None,
                "localEvaluators": evaluators_config if evaluators_config else [],
            }
            response = self._client.post(url, json=payload)
            response.raise_for_status()
            data = response.json()
            if isinstance(data, dict) and "data" in data:
                return data.get("data", {})
        except Exception:
            response_json = response.json()
            logger.error(
                "netra.evaluation: Failed to create run '%s': %s", name, response_json.get("error").get("message", "")
            )
            return {"success": False}

    def post_run_item(self, run_id: str, payload: Dict[str, Any]) -> Any:
        """
        Submit a new run item to the backend.

        Args:
            run_id: The id of the run to which the item will be added.
            payload: The run item to add.

        Returns:
            A backend JSON response on success or {"success": False} on error.
        """
        if not self._client:
            logger.error("netra.evaluation: Evaluation client is not initialized; cannot post run item")
            return {"success": False}
        try:
            url = f"/evaluations/run/{run_id}/item"
            response = self._client.post(url, json=payload)
            response.raise_for_status()
            data = response.json()
            if isinstance(data, dict) and "data" in data:
                run_item = data.get("data", {}).get("item")
                run_item_id = run_item.get("id")
                return run_item_id
            return data
        except Exception:
            response_json = response.json()
            logger.error(
                "netra.evaluation: Failed to post run item for run '%s': %s",
                run_id,
                response_json.get("error").get("message", ""),
            )
            return {"success": False}

    def submit_local_evaluations(
        self, run_id: str, test_run_item_id: str, evaluator_results: List[Dict[str, Any]]
    ) -> Any:
        """
        Submit local evaluations result

        Args:
            run_id: The id of the run to which the item will be added.
            test_run_item_id: The id of the test run item.
            evaluator_results: The evaluator results to submit.

        Returns:
            A backend JSON response containing confirmation of the submission.
        """
        if not self._client:
            logger.error("netra.evaluation: Evaluation client is not initialized; cannot submit local evaluations")
            return {"success": False}
        try:
            url = f"/evaluations/run/{run_id}/item/{test_run_item_id}/local-evaluations"
            payload: Dict[str, Any] = {"evaluatorResults": evaluator_results}
            response = self._client.post(url, json=payload)
            response.raise_for_status()
            data = response.json()
            if isinstance(data, dict) and "data" in data:
                return data.get("data", {})
            return data
        except Exception:
            response_json = response.json()
            logger.error(
                "netra.evaluation: Failed to submit local evaluations for run '%s', item '%s': %s",
                run_id,
                test_run_item_id,
                response_json.get("error").get("message", ""),
            )
            return {"success": False}

    def post_run_status(self, run_id: str, status: str) -> Any:
        """
        Submit the run status

        Args:
             run_id: The id of the run to which the item will be added.
             status: The status of the run.

         Returns:
             A backend JSON response containing confirmation of the submission.
        """
        if not self._client:
            logger.error("netra.evaluation: Evaluation client is not initialized; cannot post run status")
            return {"success": False}
        try:
            url = f"/evaluations/run/{run_id}/status"
            payload: Dict[str, Any] = {"status": status}
            response = self._client.post(url, json=payload)
            response.raise_for_status()
            data = response.json()
            if isinstance(data, dict) and "data" in data:
                logger.info("netra.evaluation: Completed test run successfully")
                return data.get("data", {})
            return data
        except Exception:
            response_json = response.json()
            logger.error(
                "netra.evaluation: Failed to post run status for run '%s': %s",
                run_id,
                response_json.get("error").get("message", ""),
            )
            return {"success": False}

    def get_span_by_id(self, span_id: str) -> Any:
        """
        Check if a span exists in the backend.

        Args:
            span_id: The span ID to check.

        Returns:
            The span data if found, None otherwise.
        """
        if not self._client:
            logger.error("netra.evaluation: Evaluation client is not initialized; cannot get span")
            return None
        try:
            url = f"sdk/traces/spans/{span_id}"
            response = self._client.get(url)
            response.raise_for_status()
            data = response.json()
            if isinstance(data, dict):
                return data.get("data", data)
            return data
        except Exception:
            return None

    async def wait_for_span_ingestion(
        self,
        span_id: str,
        timeout_seconds: float = 60.0,
        poll_interval_seconds: float = 1.0,
        initial_delay_seconds: float = 0.5,
    ) -> bool:
        """
        Wait until a span is available in the backend.

        Polls the GET /spans/:id endpoint to verify span availability
        before running evaluators.

        Args:
            span_id: The span ID to poll for.
            timeout_seconds: Maximum time to wait for span ingestion.
            poll_interval_seconds: Time between polling attempts.
            initial_delay_seconds: Initial delay before first poll attempt.

        Returns:
            True if span was found within timeout, False otherwise.
        """
        if not span_id:
            return False

        await asyncio.sleep(initial_delay_seconds)

        deadline = time.time() + timeout_seconds
        while time.time() < deadline:
            span_data = await asyncio.to_thread(self.get_span_by_id, span_id)
            if span_data is not None:
                return True
            await asyncio.sleep(poll_interval_seconds)

        return False



import logging
from typing import Any

from opentelemetry import context as context_api
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.semconv._incubating.attributes.gen_ai_attributes import GEN_AI_RESPONSE_ID
from opentelemetry.semconv_ai import SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY, LLMRequestTypeValues, SpanAttributes
from opentelemetry.trace import Span

logger = logging.getLogger(__name__)

GEN_AI_SYSTEM = "gen_ai.system"
GEN_AI_SYSTEM_CEREBRAS = "cerebras"


def model_as_dict(input_object: Any) -> Any:
    """Convert OpenAI model object to dictionary"""
    if hasattr(input_object, "model_dump"):
        return input_object.model_dump()

    elif hasattr(input_object, "to_dict"):
        return input_object.to_dict()

    elif isinstance(input_object, dict):
        return input_object

    else:
        return {}


def should_suppress_instrumentation() -> bool:
    """Check if instrumentation should be suppressed via OTel context keys."""
    return bool(
        context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY)
        or context_api.get_value(SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY)
    )


def set_request_attributes(span: Span, llm_request_type: LLMRequestTypeValues, kwargs: dict[Any, Any]) -> None:
    """Set request attributes on the span based on Cerebras request kwargs."""
    if not span.is_recording():
        return

    if model := kwargs.get("model"):
        span.set_attribute(SpanAttributes.LLM_REQUEST_MODEL, model)
    if max_tokens := kwargs.get("max_tokens"):
        span.set_attribute(SpanAttributes.LLM_REQUEST_MAX_TOKENS, max_tokens)
    if temperature := kwargs.get("temperature"):
        span.set_attribute(SpanAttributes.LLM_REQUEST_TEMPERATURE, temperature)
    if top_p := kwargs.get("top_p"):
        span.set_attribute(SpanAttributes.LLM_REQUEST_TOP_P, top_p)
    if frequency_penalty := kwargs.get("frequency_penalty"):
        span.set_attribute(SpanAttributes.LLM_FREQUENCY_PENALTY, frequency_penalty)
    if presence_penalty := kwargs.get("presence_penalty"):
        span.set_attribute(SpanAttributes.LLM_PRESENCE_PENALTY, presence_penalty)

    if llm_request_type == LLMRequestTypeValues.COMPLETION:
        if prompt := kwargs.get("prompt"):
            span.set_attribute(f"{SpanAttributes.LLM_PROMPTS}.0.role", "user")
            span.set_attribute(f"{SpanAttributes.LLM_PROMPTS}.0.content", prompt)
    elif llm_request_type == LLMRequestTypeValues.CHAT:
        if messages := kwargs.get("messages"):
            if isinstance(messages, list):
                for index, message in enumerate(messages):
                    if isinstance(message, dict):
                        span.set_attribute(
                            f"{SpanAttributes.LLM_PROMPTS}.{index}.role",
                            message.get("role", "user"),
                        )
                        span.set_attribute(
                            f"{SpanAttributes.LLM_PROMPTS}.{index}.content",
                            message.get("content"),
                        )
                    elif (role := message.role) and (content := message.content):
                        span.set_attribute(
                            f"{SpanAttributes.LLM_PROMPTS}.{index}.role",
                            role,
                        )
                        span.set_attribute(
                            f"{SpanAttributes.LLM_PROMPTS}.{index}.content",
                            content,
                        )


def set_response_attributes(span: Span, response: Any) -> None:
    """Set response-related attributes, always capturing usage and cached tokens."""
    if not span.is_recording():
        return

    response_dict = model_as_dict(response)

    try:
        if response_dict.get("usage"):
            usage = response_dict.get("usage")
            if usage.get("prompt_tokens"):
                span.set_attribute(SpanAttributes.LLM_USAGE_PROMPT_TOKENS, usage.get("prompt_tokens"))
            if usage.get("completion_tokens"):
                span.set_attribute(SpanAttributes.LLM_USAGE_COMPLETION_TOKENS, usage.get("completion_tokens"))
            if usage.get("total_tokens"):
                span.set_attribute(SpanAttributes.LLM_USAGE_TOTAL_TOKENS, usage.get("total_tokens"))
            if usage.get("prompt_tokens_details"):
                details = usage.get("prompt_tokens_details", None)
                if details and hasattr(details, "cached_tokens"):
                    span.set_attribute(SpanAttributes.LLM_USAGE_CACHE_READ_INPUT_TOKENS, details.cached_tokens)
                elif isinstance(details, dict) and "cached_tokens" in details:
                    span.set_attribute(SpanAttributes.LLM_USAGE_CACHE_READ_INPUT_TOKENS, details["cached_tokens"])

        if response_dict.get("id"):
            span.set_attribute(GEN_AI_RESPONSE_ID, response_dict.get("id"))
        if response_dict.get("choices") and response_dict.get("choices"):
            for index, choice in enumerate(response_dict.get("choices")):
                prefix = f"{SpanAttributes.LLM_COMPLETIONS}.{index}"
                if choice.get("message"):
                    span.set_attribute(f"{prefix}.role", choice.get("message", {}).get("role", "assistant"))
                    span.set_attribute(f"{prefix}.content", choice.get("message", {}).get("content", ""))
                if choice.get("text"):
                    span.set_attribute(f"{prefix}.role", "assistant")
                    span.set_attribute(f"{prefix}.content", choice.get("text"))
                if choice.get("finish_reason"):
                    span.set_attribute(f"{prefix}.finish_reason", choice.get("finish_reason"))

    except Exception:
        logger.exception("Failed to set response attributes")



from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class SessionUsageData(BaseModel):  # type:ignore[misc]
    session_id: str
    token_count: int
    request_count: int
    total_cost: float


class TenantUsageData(BaseModel):  # type:ignore[misc]
    tenant_id: str
    organisation_id: str
    token_count: int
    request_count: int
    session_count: int
    total_cost: float


class TraceSummary(BaseModel):  # type:ignore[misc]
    id: str
    name: str
    kind: str
    latency_ms: int
    start_time: str
    end_time: str
    cursor: str
    organisation_id: str
    project_id: str
    session_id: Optional[str] = None
    tenant_id: Optional[str] = None
    user_id: Optional[str] = None
    environment: Optional[str] = None
    models: Optional[List[str]] = None
    prompt_tokens: Optional[int] = None
    completion_tokens: Optional[int] = None
    cached_tokens: Optional[int] = None
    total_tokens: Optional[int] = None
    prompt_tokens_cost: Optional[float] = None
    completion_tokens_cost: Optional[float] = None
    cached_tokens_cost: Optional[float] = None
    total_cost: Optional[float] = None
    has_pii: Optional[bool] = None
    pii_entities: Optional[List[Any]] = None
    has_violation: Optional[bool] = None
    violations: Optional[List[Any]] = None
    has_error: Optional[bool] = None
    service: Optional[str] = None


class TracesPage(BaseModel):  # type:ignore[misc]
    traces: List[TraceSummary]
    has_next_page: bool
    next_cursor: Optional[str] = None


class TraceSpan(BaseModel):  # type:ignore[misc]
    id: str
    trace_id: str
    name: str
    kind: str
    parent_span_id: Optional[str] = None
    latency_ms: Optional[int] = None
    start_time_ms: Optional[str] = None
    end_time_ms: Optional[str] = None
    organisation_id: str
    project_id: str
    session_id: Optional[str] = None
    tenant_id: Optional[str] = None
    user_id: Optional[str] = None
    environment: Optional[str] = None
    model_name: Optional[str] = None
    prompt_tokens: Optional[int] = None
    completion_tokens: Optional[int] = None
    cached_tokens: Optional[int] = None
    total_tokens: Optional[int] = None
    prompt_tokens_cost: Optional[float] = None
    completion_tokens_cost: Optional[float] = None
    cached_tokens_cost: Optional[float] = None
    total_cost: Optional[float] = None
    has_pii: Optional[bool] = None
    pii_entities: Optional[List[Any]] = None
    pii_actions: Optional[Dict[str, Any]] = None
    has_violation: Optional[bool] = None
    violations: Optional[List[Any]] = None
    violation_actions: Optional[Dict[str, Any]] = None
    status: Optional[str] = None
    attributes: Optional[str] = None
    events: Optional[str] = None
    links: Optional[str] = None
    resources: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    has_error: Optional[bool] = None
    created_at: str
    cursor: str


class SpansPage(BaseModel):  # type:ignore[misc]
    spans: List[TraceSpan]
    has_next_page: bool
    next_cursor: Optional[str] = None



"""HTTP client for simulation API endpoints."""

import logging
import os
from typing import Any, Optional

import httpx

from netra.config import Config
from netra.simulation.models import ConversationResponse, SimulationItem

logger = logging.getLogger(__name__)

_DEFAULT_TIMEOUT = 10.0
_LOG_PREFIX = "netra.simulation"


class SimulationHttpClient:
    """Internal HTTP client for simulation API endpoints.

    Attributes:
        _client: The underlying httpx client instance.
    """

    __slots__ = ("_client",)

    def __init__(self, config: Config) -> None:
        """Initialize the simulation HTTP client.

        Args:
            config: The Netra configuration object.
        """
        self._client: Optional[httpx.Client] = self._create_client(config)

    def _create_client(self, config: Config) -> Optional[httpx.Client]:
        """Create and configure the HTTP client.

        Args:
            config: The Netra configuration object.

        Returns:
            Configured httpx client or None if creation fails.
        """
        endpoint = (config.otlp_endpoint or "").strip()
        if not endpoint:
            logger.error("%s: NETRA_OTLP_ENDPOINT is required", _LOG_PREFIX)
            return None

        base_url = self._resolve_base_url(endpoint)
        headers = self._build_headers(config)
        timeout = self._get_timeout()

        try:
            return httpx.Client(base_url=base_url, headers=headers, timeout=timeout)
        except Exception as exc:
            logger.error("%s: Failed to create HTTP client: %s", _LOG_PREFIX, exc)
            return None

    def _resolve_base_url(self, endpoint: str) -> str:
        """Extract base URL, removing telemetry suffix if present.

        Args:
            endpoint: The raw endpoint URL.

        Returns:
            The cleaned base URL.
        """
        base_url = endpoint.rstrip("/")
        if base_url.endswith("/telemetry"):
            base_url = base_url[: -len("/telemetry")]
        return base_url

    def _build_headers(self, config: Config) -> dict[str, str]:
        """Build request headers from configuration.

        Args:
            config: The Netra configuration object.

        Returns:
            Dictionary of HTTP headers.
        """
        headers: dict[str, str] = dict(config.headers or {})
        if config.api_key:
            headers["x-api-key"] = config.api_key
        return headers

    def _get_timeout(self) -> float:
        """Get timeout from environment or use default.

        Returns:
            The timeout value in seconds.
        """
        timeout_str = os.getenv("NETRA_SIMULATION_TIMEOUT")
        if not timeout_str:
            return _DEFAULT_TIMEOUT
        try:
            return float(timeout_str)
        except ValueError:
            logger.warning(
                "%s: Invalid timeout '%s', using default %.1f",
                _LOG_PREFIX,
                timeout_str,
                _DEFAULT_TIMEOUT,
            )
            return _DEFAULT_TIMEOUT

    def create_run(
        self,
        name: str,
        dataset_id: str,
        context: Optional[dict[str, Any]] = None,
    ) -> Optional[dict[str, Any]]:
        """Create a new simulation run for the specified dataset.

        Args:
            name: Name of the simulation run.
            dataset_id: Identifier of the dataset to simulate.
            context: Optional context data for the simulation.

        Returns:
            Dictionary containing run_id and simulation_items, or None on failure.
        """
        if not self._client:
            logger.error("%s: Client not initialized", _LOG_PREFIX)
            return None

        response: Optional[httpx.Response] = None
        try:
            url = "/evaluations/test_run/multi-turn"
            payload: dict[str, Any] = {
                "name": name,
                "datasetId": dataset_id,
                "context": context or {},
            }
            response = self._client.post(url, json=payload, timeout=500)
            response.raise_for_status()
            data = response.json()

            response_data = data.get("data", {})
            user_messages = response_data.get("userMessages", [])
            if not user_messages:
                logger.warning("%s: No user messages returned from create_run", _LOG_PREFIX)
                return None

            run_id = response_data.get("id", "")
            simulation_items = [
                SimulationItem(
                    run_item_id=msg.get("testRunItemId", ""),
                    message=msg.get("userMessage", ""),
                    turn_id=msg.get("turnId", ""),
                )
                for msg in user_messages
            ]
            return {
                "run_id": run_id,
                "simulation_items": simulation_items,
            }

        except Exception as exc:
            error_msg = self._extract_error_message(response, exc)
            logger.error("%s: Failed to create simulation run: %s", _LOG_PREFIX, error_msg)
            return None

    def trigger_conversation(
        self,
        message: str,
        turn_id: str,
        session_id: str,
        trace_id: str,
    ) -> Optional[ConversationResponse]:
        """Send a conversation turn to the backend and get the next response.

        Args:
            message: Agent response message.
            turn_id: Turn identifier.
            session_id: Session identifier.
            trace_id: Trace identifier.

        Returns:
            ConversationResponse with next turn info, or None on failure.
        """
        if not self._client:
            logger.error("%s: Client not initialized", _LOG_PREFIX)
            return None

        response: Optional[httpx.Response] = None
        try:
            url = "/evaluations/turn/agent-response"
            payload: dict[str, Any] = {
                "turnId": turn_id,
                "agentResponse": {"message": message},
                "sessionId": session_id,
                "traceId": trace_id,
            }

            response = self._client.post(url, json=payload, timeout=500)
            response.raise_for_status()
            data = response.json()

            response_data = data.get("data", {})
            decision = response_data.get("decision", "continue")

            if decision == "stop":
                return ConversationResponse(
                    decision=decision,
                    reason=response_data.get("reason", ""),
                )

            user_messages = response_data.get("userMessages", [])
            if not user_messages:
                logger.warning("%s: No user messages in continue response", _LOG_PREFIX)
                return None

            next_msg = next(iter(user_messages))
            return ConversationResponse(
                decision=decision,
                next_turn_id=next_msg.get("turnId", ""),
                next_user_message=next_msg.get("userMessage", ""),
                next_run_item_id=next_msg.get("testRunItemId", ""),
            )

        except Exception as exc:
            error_msg = self._extract_error_message(response, exc)
            logger.error("%s: Failed to trigger conversation: %s", _LOG_PREFIX, error_msg)
            raise

    def report_failure(self, run_id: str, run_item_id: str, error: str) -> None:
        """Report a task execution failure to the backend.

        Args:
            run_id: Identifier of the run.
            run_item_id: Identifier of the run item.
            error: Error message describing the failure.
        """
        if not self._client:
            logger.error("%s: Client not initialized", _LOG_PREFIX)
            return

        response: Optional[httpx.Response] = None
        try:
            url = f"/evaluations/run/{run_id}/item/{run_item_id}/status"
            payload: dict[str, Any] = {"status": "failed", "failureReason": error}
            response = self._client.patch(url, json=payload)
            response.raise_for_status()
            logger.info("%s: Reported failure - %s", _LOG_PREFIX, error)
        except Exception as exc:
            error_msg = self._extract_error_message(response, exc)
            logger.error("%s: Failed to report failure: %s", _LOG_PREFIX, error_msg)

    def post_run_status(self, run_id: str, status: str) -> Any:
        """Submit the run status.

        Args:
            run_id: The id of the run to update.
            status: The status of the run.

        Returns:
            Backend JSON response containing confirmation, or error dict.
        """
        if not self._client:
            logger.error("%s: Client not initialized; cannot post run status", _LOG_PREFIX)
            return {"success": False}

        response: Optional[httpx.Response] = None
        try:
            url = f"/evaluations/run/{run_id}/status"
            payload: dict[str, Any] = {"status": status}
            response = self._client.post(url, json=payload)
            response.raise_for_status()
            data = response.json()
            if isinstance(data, dict) and "data" in data:
                logger.info("%s: Test run status %s", _LOG_PREFIX, status)
                return data.get("data", {})
            return data
        except Exception as exc:
            error_msg = self._extract_error_message(response, exc)
            logger.error("%s: Failed to post run status for run '%s': %s", _LOG_PREFIX, run_id, error_msg)
            return {"success": False}

    def _extract_error_message(
        self,
        response: Optional[httpx.Response],
        exc: Exception,
    ) -> Any:
        """Extract error message from response or exception.

        Args:
            response: The HTTP response object, if available.
            exc: The exception that was raised.

        Returns:
            A descriptive error message string.
        """
        if response is not None:
            try:
                response_json = response.json()
                error_data = response_json.get("error", {})
                if isinstance(error_data, dict):
                    return error_data.get("message", str(exc))
            except Exception:
                pass
        return str(exc)



from typing import Any, Dict, Optional

from opentelemetry import context as context_api
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.semconv_ai import SpanAttributes
from opentelemetry.trace.status import Status, StatusCode

# Constants for consistent truncation and limits
MAX_CONTENT_LENGTH = 1000
MAX_ARGS_LENGTH = 500
MAX_ITEMS_TO_PROCESS = 5
MAX_FUNCTIONS_TO_PROCESS = 3


def _safe_set_attribute(span: Any, key: str, value: Any, max_length: Optional[int] = None) -> None:
    """Safely set span attribute with optional truncation and null checks."""
    if not span.is_recording() or value is None:
        return

    str_value = str(value)
    if max_length and len(str_value) > max_length:
        str_value = str_value[:max_length]

    span.set_attribute(key, str_value)


def _safe_get_attribute(obj: Any, attr_name: str, default: Any = None) -> Any:
    """Safely get attribute from object with default fallback."""
    return getattr(obj, attr_name, default) if hasattr(obj, attr_name) else default


def _handle_span_error(span: Any, exception: Exception) -> None:
    """Common error handling for spans."""
    span.set_status(Status(StatusCode.ERROR, str(exception)))
    _safe_set_attribute(span, "error.type", type(exception).__name__)
    _safe_set_attribute(span, "error.message", str(exception))


def _set_timing_attributes(span: Any, start_time: float, end_time: float) -> None:
    """Set timing attributes on span."""
    duration_ms = (end_time - start_time) * 1000
    _safe_set_attribute(span, "llm.response.duration", duration_ms)


def _set_assistant_response_content(span: Any, result: Any, finish_reason: str = "completed") -> None:
    """Set assistant response content in OpenAI wrapper format."""
    if not span.is_recording():
        return

    # Set the assistant response in the same format as OpenAI wrapper
    index = 0  # Always use index 0 for pydantic_ai responses
    _safe_set_attribute(span, f"{SpanAttributes.LLM_COMPLETIONS}.{index}.role", "assistant")
    _safe_set_attribute(span, f"{SpanAttributes.LLM_COMPLETIONS}.{index}.finish_reason", finish_reason)

    # Get the output content from the result
    output = _safe_get_attribute(result, "output")
    if output is not None:
        _safe_set_attribute(span, f"{SpanAttributes.LLM_COMPLETIONS}.{index}.content", output, MAX_CONTENT_LENGTH)


def set_pydantic_request_attributes(
    span: Any,
    kwargs: Dict[str, Any],
    operation_type: str,
    model_name: Optional[str] = None,
    include_model: bool = False,
) -> None:
    """Set request attributes on span for pydantic_ai."""
    if not span.is_recording():
        return

    # Set operation type
    _safe_set_attribute(span, f"{SpanAttributes.LLM_REQUEST_TYPE}", operation_type)

    # Set model only if explicitly requested (for CallToolsNode spans)
    if include_model and model_name:
        _safe_set_attribute(span, f"{SpanAttributes.LLM_REQUEST_MODEL}", model_name)

    # Set temperature and max_tokens if available
    _safe_set_attribute(span, f"{SpanAttributes.LLM_REQUEST_TEMPERATURE}", kwargs.get("temperature"))
    _safe_set_attribute(span, f"{SpanAttributes.LLM_REQUEST_MAX_TOKENS}", kwargs.get("max_tokens"))


def set_pydantic_response_attributes(span: Any, result: Any) -> None:
    """Set response attributes on span for pydantic_ai."""
    if not span.is_recording():
        return

    # Set response model if available
    model_name = _safe_get_attribute(result, "model_name")
    _safe_set_attribute(span, f"{SpanAttributes.LLM_RESPONSE_MODEL}", model_name)

    # Set usage information
    if hasattr(result, "usage"):
        usage = result.usage()
        if usage:
            _safe_set_attribute(
                span, f"{SpanAttributes.LLM_USAGE_PROMPT_TOKENS}", _safe_get_attribute(usage, "request_tokens")
            )
            _safe_set_attribute(
                span, f"{SpanAttributes.LLM_USAGE_COMPLETION_TOKENS}", _safe_get_attribute(usage, "response_tokens")
            )
            _safe_set_attribute(
                span, f"{SpanAttributes.LLM_USAGE_TOTAL_TOKENS}", _safe_get_attribute(usage, "total_tokens")
            )

            # Set any additional details from usage
            details = _safe_get_attribute(usage, "details")
            if details:
                for key, value in details.items():
                    if value:
                        _safe_set_attribute(span, f"gen_ai.usage.details.{key}", value)

    # Set output content if available
    output = _safe_get_attribute(result, "output")
    if output is not None:
        _safe_set_attribute(span, f"{SpanAttributes.LLM_COMPLETIONS}.0.role", "assistant")
        _safe_set_attribute(span, f"{SpanAttributes.LLM_COMPLETIONS}.0.content", output, MAX_CONTENT_LENGTH)


def should_suppress_instrumentation() -> bool:
    """Check if instrumentation should be suppressed"""
    return context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY) is True


def get_node_span_name(node: Any) -> str:
    """Get appropriate span name for a node type"""
    node_type = type(node).__name__
    if "UserPrompt" in node_type:
        return "pydantic_ai.node.user_prompt"
    elif "ModelRequest" in node_type:
        return "pydantic_ai.node.model_request"
    elif "CallTools" in node_type:
        return "pydantic_ai.node.call_tools"
    elif "End" in node_type:
        return "pydantic_ai.node.end"
    else:
        return f"pydantic_ai.node.{node_type.lower()}"


def set_node_attributes(span: Any, node: Any) -> None:
    """Set attributes on span based on node type and content"""
    if not span.is_recording():
        return

    node_type = type(node).__name__
    _safe_set_attribute(span, "pydantic_ai.node.type", node_type)

    # UserPromptNode attributes
    if "UserPrompt" in node_type:
        _set_user_prompt_node_attributes(span, node)

    # ModelRequestNode attributes
    elif "ModelRequest" in node_type:
        _set_model_request_node_attributes(span, node)

    # CallToolsNode attributes
    elif "CallTools" in node_type:
        _set_call_tools_node_attributes(span, node)

    # End node attributes
    elif "End" in node_type:
        _set_end_node_attributes(span, node)

    # Generic node attributes for any other node types
    else:
        _set_generic_node_attributes(span, node)


def _set_user_prompt_node_attributes(span: Any, node: Any) -> None:
    """Set attributes specific to UserPromptNode."""
    # User prompt content
    user_prompt = _safe_get_attribute(node, "user_prompt")
    if user_prompt:
        _safe_set_attribute(span, f"{SpanAttributes.LLM_PROMPTS}.0.role", "user")
        _safe_set_attribute(span, f"{SpanAttributes.LLM_PROMPTS}.0.content", user_prompt, MAX_CONTENT_LENGTH)
        _safe_set_attribute(span, "pydantic_ai.user_prompt", user_prompt, MAX_CONTENT_LENGTH)

    # Instructions
    instructions = _safe_get_attribute(node, "instructions")
    _safe_set_attribute(span, "pydantic_ai.instructions", instructions, MAX_CONTENT_LENGTH)

    # Instructions functions
    instructions_functions = _safe_get_attribute(node, "instructions_functions")
    if instructions_functions:
        _safe_set_attribute(span, "pydantic_ai.instructions_functions_count", len(instructions_functions))
        for i, func in enumerate(instructions_functions[:MAX_FUNCTIONS_TO_PROCESS]):
            func_name = _safe_get_attribute(func, "__name__")
            _safe_set_attribute(span, f"pydantic_ai.instructions_functions.{i}.name", func_name)

    # System prompts
    system_prompts = _safe_get_attribute(node, "system_prompts")
    if system_prompts:
        _safe_set_attribute(span, "pydantic_ai.system_prompts_count", len(system_prompts))
        for i, prompt in enumerate(system_prompts[:MAX_FUNCTIONS_TO_PROCESS]):
            _safe_set_attribute(span, f"pydantic_ai.system_prompts.{i}", prompt, MAX_ARGS_LENGTH)

    # System prompt functions
    system_prompt_functions = _safe_get_attribute(node, "system_prompt_functions")
    if system_prompt_functions:
        _safe_set_attribute(span, "pydantic_ai.system_prompt_functions_count", len(system_prompt_functions))

    # System prompt dynamic functions
    system_prompt_dynamic_functions = _safe_get_attribute(node, "system_prompt_dynamic_functions")
    if system_prompt_dynamic_functions:
        _safe_set_attribute(
            span, "pydantic_ai.system_prompt_dynamic_functions_count", len(system_prompt_dynamic_functions)
        )
        for key in list(system_prompt_dynamic_functions.keys())[:MAX_FUNCTIONS_TO_PROCESS]:
            func_type = type(system_prompt_dynamic_functions[key]).__name__
            _safe_set_attribute(span, f"pydantic_ai.system_prompt_dynamic_functions.{key}", func_type)


def _set_model_request_node_attributes(span: Any, node: Any) -> None:
    """Set attributes specific to ModelRequestNode."""
    request = _safe_get_attribute(node, "request")
    if not request:
        return

    # Request parts
    parts = _safe_get_attribute(request, "parts")
    if parts:
        _safe_set_attribute(span, "pydantic_ai.request.parts_count", len(parts))

        for i, part in enumerate(parts[:MAX_ITEMS_TO_PROCESS]):
            part_type = type(part).__name__
            _safe_set_attribute(span, f"pydantic_ai.request.parts.{i}.type", part_type)

            # Content for text parts
            content = _safe_get_attribute(part, "content")
            if content:
                _safe_set_attribute(span, f"{SpanAttributes.LLM_PROMPTS}.{i}.content", content, MAX_CONTENT_LENGTH)
                _safe_set_attribute(span, f"pydantic_ai.request.parts.{i}.content", content, MAX_CONTENT_LENGTH)

            # Role for message parts
            role = _safe_get_attribute(part, "role")
            if role:
                _safe_set_attribute(span, f"{SpanAttributes.LLM_PROMPTS}.{i}.role", role)
                _safe_set_attribute(span, f"pydantic_ai.request.parts.{i}.role", role)

            # Timestamp, tool call information
            _safe_set_attribute(
                span, f"pydantic_ai.request.parts.{i}.timestamp", _safe_get_attribute(part, "timestamp")
            )
            _safe_set_attribute(
                span, f"pydantic_ai.request.parts.{i}.tool_name", _safe_get_attribute(part, "tool_name")
            )
            _safe_set_attribute(
                span, f"pydantic_ai.request.parts.{i}.tool_call_id", _safe_get_attribute(part, "tool_call_id")
            )
            _safe_set_attribute(
                span, f"pydantic_ai.request.parts.{i}.args", _safe_get_attribute(part, "args"), MAX_ARGS_LENGTH
            )

    # Request metadata
    _safe_set_attribute(span, "pydantic_ai.request.model_name", _safe_get_attribute(request, "model_name"))
    _safe_set_attribute(span, f"{SpanAttributes.LLM_REQUEST_TEMPERATURE}", _safe_get_attribute(request, "temperature"))
    _safe_set_attribute(span, f"{SpanAttributes.LLM_REQUEST_MAX_TOKENS}", _safe_get_attribute(request, "max_tokens"))


def _set_call_tools_node_attributes(span: Any, node: Any) -> None:
    """Set attributes specific to CallToolsNode."""
    response = _safe_get_attribute(node, "model_response")
    if not response:
        return

    # Response parts
    parts = _safe_get_attribute(response, "parts")
    if parts:
        _safe_set_attribute(span, "pydantic_ai.response.parts_count", len(parts))

        for i, part in enumerate(parts[:MAX_ITEMS_TO_PROCESS]):
            part_type = type(part).__name__
            _safe_set_attribute(span, f"pydantic_ai.response.parts.{i}.type", part_type)

            # Content for text parts
            content = _safe_get_attribute(part, "content")
            if content:
                _safe_set_attribute(span, f"{SpanAttributes.LLM_COMPLETIONS}.{i}.content", content, MAX_CONTENT_LENGTH)
                _safe_set_attribute(span, f"pydantic_ai.response.parts.{i}.content", content, MAX_CONTENT_LENGTH)
                _safe_set_attribute(span, f"{SpanAttributes.LLM_COMPLETIONS}.{i}.role", "assistant")

            # Tool call information
            _safe_set_attribute(
                span, f"pydantic_ai.response.parts.{i}.tool_name", _safe_get_attribute(part, "tool_name")
            )
            _safe_set_attribute(
                span, f"pydantic_ai.response.parts.{i}.tool_call_id", _safe_get_attribute(part, "tool_call_id")
            )
            _safe_set_attribute(
                span, f"pydantic_ai.response.parts.{i}.args", _safe_get_attribute(part, "args"), MAX_ARGS_LENGTH
            )

    # Usage information
    usage = _safe_get_attribute(response, "usage")
    if usage:
        _safe_set_attribute(span, "pydantic_ai.usage.requests", _safe_get_attribute(usage, "requests"))

        # Token usage with dual attributes for compatibility
        request_tokens = _safe_get_attribute(usage, "request_tokens")
        if request_tokens is not None:
            _safe_set_attribute(span, f"{SpanAttributes.LLM_USAGE_PROMPT_TOKENS}", request_tokens)
            _safe_set_attribute(span, "pydantic_ai.usage.request_tokens", request_tokens)

        response_tokens = _safe_get_attribute(usage, "response_tokens")
        if response_tokens is not None:
            _safe_set_attribute(span, f"{SpanAttributes.LLM_USAGE_COMPLETION_TOKENS}", response_tokens)
            _safe_set_attribute(span, "pydantic_ai.usage.response_tokens", response_tokens)

        total_tokens = _safe_get_attribute(usage, "total_tokens")
        if total_tokens is not None:
            _safe_set_attribute(span, f"{SpanAttributes.LLM_USAGE_TOTAL_TOKENS}", total_tokens)
            _safe_set_attribute(span, "pydantic_ai.usage.total_tokens", total_tokens)

        # Additional usage details
        details = _safe_get_attribute(usage, "details")
        if details:
            for key, value in details.items():
                if value is not None:
                    _safe_set_attribute(span, f"pydantic_ai.usage.details.{key}", value)

    # Model information (only for CallToolsNode)
    model_name = _safe_get_attribute(response, "model_name")
    if model_name:
        _safe_set_attribute(span, f"{SpanAttributes.LLM_RESPONSE_MODEL}", model_name)
        _safe_set_attribute(span, f"{SpanAttributes.LLM_REQUEST_MODEL}", model_name)
        _safe_set_attribute(span, "pydantic_ai.response.model_name", model_name)

    # Timestamp
    _safe_set_attribute(span, "pydantic_ai.response.timestamp", _safe_get_attribute(response, "timestamp"))

    # Tool execution results (if available)
    tool_results = _safe_get_attribute(node, "tool_results")
    if tool_results:
        _safe_set_attribute(span, "pydantic_ai.tool_results_count", len(tool_results))
        for i, result in enumerate(tool_results[:MAX_FUNCTIONS_TO_PROCESS]):
            _safe_set_attribute(span, f"pydantic_ai.tool_results.{i}", result, MAX_ARGS_LENGTH)


def _set_end_node_attributes(span: Any, node: Any) -> None:
    """Set attributes specific to End node."""
    data = _safe_get_attribute(node, "data")
    if not data:
        return

    # Final output
    output = _safe_get_attribute(data, "output")
    if output is not None:
        _safe_set_attribute(span, f"{SpanAttributes.LLM_COMPLETIONS}.0.role", "assistant")
        _safe_set_attribute(span, f"{SpanAttributes.LLM_COMPLETIONS}.0.content", output, MAX_CONTENT_LENGTH)
        _safe_set_attribute(span, "pydantic_ai.final_output", output, MAX_CONTENT_LENGTH)

    # Cost information
    cost = _safe_get_attribute(data, "cost")
    if cost is not None:
        _safe_set_attribute(span, "pydantic_ai.cost", float(cost))

    # Usage summary
    usage = _safe_get_attribute(data, "usage")
    if usage:
        _safe_set_attribute(span, "pydantic_ai.final_usage.total_tokens", _safe_get_attribute(usage, "total_tokens"))
        _safe_set_attribute(
            span, "pydantic_ai.final_usage.request_tokens", _safe_get_attribute(usage, "request_tokens")
        )
        _safe_set_attribute(
            span, "pydantic_ai.final_usage.response_tokens", _safe_get_attribute(usage, "response_tokens")
        )

    # Messages history
    messages = _safe_get_attribute(data, "messages")
    if messages:
        _safe_set_attribute(span, "pydantic_ai.messages_count", len(messages))

    # New messages
    new_messages = _safe_get_attribute(data, "new_messages")
    if new_messages:
        _safe_set_attribute(span, "pydantic_ai.new_messages_count", len(new_messages))


def _set_generic_node_attributes(span: Any, node: Any) -> None:
    """Set attributes for any other node types."""
    # Try to extract common attributes that might be available
    for attr_name in ["content", "message", "value", "result", "error"]:
        attr_value = _safe_get_attribute(node, attr_name)
        _safe_set_attribute(span, f"pydantic_ai.{attr_name}", attr_value, MAX_CONTENT_LENGTH)



import logging
from typing import Any

from netra.config import Config
from netra.prompts.client import PromptsHttpClient

logger = logging.getLogger(__name__)


class Prompts:
    """
    Public entry-point exposed as Netra.prompts
    """

    def __init__(self, cfg: Config) -> None:
        """
        Initialize the Prompts client.

        Args:
            cfg: Configuration object containing API key and base URL
        """
        self._config = cfg
        self._client = PromptsHttpClient(cfg)

    def get_prompt(self, name: str, label: str = "production") -> Any:
        """
        Fetch a prompt version by name and label.

        Args:
            name: Name of the prompt
            label: Label of the prompt version (default: "production")

        Returns:
            Prompt version data or empty dict if not found
        """
        if not name:
            logger.error("netra.prompts: name is required to fetch a prompt")
            return None

        return self._client.get_prompt_version(prompt_name=name, label=label)



import logging
from typing import Any, Callable, Optional, Sequence

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import (
    SpanExporter,
    SpanExportResult,
)

from netra.exporters.utils import set_trial_blocked

logger = logging.getLogger(__name__)


class ResponseInterceptor:
    """Wraps requests.Session to intercept and monitor HTTP responses."""

    def __init__(self, session: Any, callback: Callable[[Any], None]) -> None:
        """Initialize the interceptor.

        Args:
            session: The requests.Session to wrap
            callback: Callable invoked with each response
        """
        self._session = session
        self._callback = callback
        self._original_request = session.request

        # Replace the request method with our wrapped version
        session.request = self._wrapped_request

    def _wrapped_request(self, *args: Any, **kwargs: Any) -> Any:
        """Wrapper around session.request that intercepts responses."""
        try:
            response = self._original_request(*args, **kwargs)
            try:
                self._callback(response)
            except Exception as e:
                logger.debug("Error in response callback: %s", e)
            return response
        except Exception:
            raise


class TrialAwareOTLPExporter(SpanExporter):  # type: ignore[misc]
    """Wrapper around OTLPSpanExporter that detects trial/quota blocks from backend."""

    def __init__(self, wrapped_exporter: SpanExporter) -> None:
        """Initialize with the exporter to wrap.

        Args:
            wrapped_exporter: The actual OTLPSpanExporter instance to wrap
        """
        self._exporter = wrapped_exporter
        self._interceptor: Optional[ResponseInterceptor] = None
        self._setup_response_interception()

    def _setup_response_interception(self) -> None:
        """Setup interception of HTTP responses from the OTLP exporter.

        This wraps the internal requests.Session to monitor responses.
        """
        try:
            # Access the internal requests.Session from OTLPSpanExporter
            if hasattr(self._exporter, "_session") and self._exporter._session is not None:
                session = self._exporter._session
                self._interceptor = ResponseInterceptor(session, self._on_http_response)
                logger.debug("Successfully setup HTTP response interception")
            else:
                logger.debug("Could not find _session on OTLPSpanExporter")
        except Exception as e:
            logger.debug("Failed to setup response interception: %s", e)

    def _on_http_response(self, response: Any) -> None:
        """Callback invoked when HTTP response is received.

        Args:
            response: The requests.Response object
        """
        try:
            self._check_response_for_trial_blocking(response)
        except Exception as e:
            logger.debug("Error processing response in callback: %s", e)

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        """Export spans and monitor response for trial status.

        Args:
            spans: Sequence of spans to export

        Returns:
            SpanExportResult from the wrapped exporter
        """
        try:
            result = self._exporter.export(spans)
            return result
        except Exception as e:
            logger.debug("Error during OTLP export: %s", e, exc_info=True)
            return SpanExportResult.FAILURE

    def _check_response_for_trial_blocking(self, response: Any) -> None:
        """Check HTTP response for trial blocking indicators.

        Args:
            response: The HTTP response object (requests.Response)
        """
        try:
            try:
                body = response.json()
            except Exception:
                body = None
            if (error := body.get("error")) and (inner := error.get("error")):
                error_code = inner.get("code")
                blocked = True if error_code == "QUOTA_EXCEEDED" else False
                if blocked:
                    logger.warning("Quota Exceeded: %s", error_code)
                    set_trial_blocked(True)
                    return
        except Exception as e:
            logger.debug("Error checking response for trial blocking: %s", e)

    def shutdown(self) -> None:
        """Shutdown the wrapped exporter."""
        try:
            self._exporter.shutdown()
        except Exception:
            pass

    def force_flush(self, timeout_millis: int = 30000) -> Any:
        """Force flush the wrapped exporter."""
        try:
            return self._exporter.force_flush(timeout_millis)
        except Exception:
            return True



import json
import logging
import threading
from typing import Any
from opentelemetry.context import Context
from opentelemetry.trace import Span, Tracer
from opentelemetry.semconv_ai import SpanAttributes
from claude_agent_sdk import (
    ClaudeAgentOptions,
    SystemMessage,
    AssistantMessage,
    UserMessage,
    ResultMessage,
    TextBlock,
    ThinkingBlock,
    ToolUseBlock,
    ToolResultBlock,
)
from netra.config import Config

logger = logging.getLogger(__name__)

# Registry to correlate ToolUseBlocks (AssistantMessage) with ToolResultBlocks (UserMessage).
# Keyed by tool_use_id; entries are removed on consumption.
_tool_call_registry: dict[str, Any] = {}
_tool_call_registry_lock = threading.Lock()


def _serialize(value: Any) -> str:
    """
    Serialize a value to a span-safe string.

    Args:
        value (Any): The value to serialize. Dicts and lists are JSON-encoded;
                     all other types are converted via str().

    Returns:
        str: A string representation of the input value.
    """
    if isinstance(value, str):
        return value
    if isinstance(value, (dict, list)):
        try:
            return json.dumps(value)
        except Exception:
            return str(value)
    return str(value)


def _set_conversation(span: Span, role: str, content: str, prompt_index: int = 0) -> int:
    """
    Write a single prompt entry to the span at the given index.

    Args:
        span (Span): The OpenTelemetry span to write attributes to.
        role (str): The conversation role (e.g. "user", "assistant", "system").
        content (str): The message content to record.
        prompt_index (int): The current index in the prompts attribute list. Defaults to 0.

    Returns:
        int: The incremented prompt index after writing.
    """
    if role and content:
        span.set_attribute(f"{SpanAttributes.LLM_PROMPTS}.{prompt_index}.role", role)
        span.set_attribute(f"{SpanAttributes.LLM_PROMPTS}.{prompt_index}.content", content)
        prompt_index += 1
    return prompt_index


def _set_usage(span: Span, usage: dict) -> None:
    """
    Write token usage attributes to the span.

    Args:
        span (Span): The span to write token counts to.
        usage (dict): A dict containing token fields such as input_tokens, output_tokens,
                      total_tokens, cache_creation_input_tokens, and cache_read_input_tokens.

    Returns:
        None
    """
    token_fields = {
        "input_tokens": SpanAttributes.LLM_USAGE_PROMPT_TOKENS,
        "output_tokens": SpanAttributes.LLM_USAGE_COMPLETION_TOKENS,
        "total_tokens": SpanAttributes.LLM_USAGE_TOTAL_TOKENS,
        "cache_creation_input_tokens": SpanAttributes.LLM_USAGE_CACHE_CREATION_INPUT_TOKENS,
        "cache_read_input_tokens": SpanAttributes.LLM_USAGE_CACHE_READ_INPUT_TOKENS,
    }
    for key, attr in token_fields.items():
        if (value := usage.get(key)) is not None:
            span.set_attribute(attr, value)


def _set_tool_result(tracer: Tracer, parent_ctx: Context, block: ToolResultBlock) -> None:
    """
    Create a child span for a tool result, correlating it with the originating tool call.

    Args:
        tracer (Tracer): The OpenTelemetry tracer used to create the child span.
        parent_ctx (Context): The parent span context to attach the child span to.
        block (ToolResultBlock): The tool result block containing the output content
                                  and the tool_use_id for correlation lookup.

    Returns:
        None
    """
    try:
        tool_call = None
        try:
            with _tool_call_registry_lock:
                tool_call = _tool_call_registry.pop(block.tool_use_id, None)
        except Exception as e:
            logger.error(f"Cannot extract tool call metadata for tool_use_id={block.tool_use_id}: {e}")

        tool_name = tool_call["name"] if tool_call else "unrecognized_tool"

        with tracer.start_as_current_span(tool_name, parent_ctx) as span:
            try:
                if tool_call:
                    span.set_attribute(f"{Config.LIBRARY_NAME}.entity.input", _serialize(tool_call.get("input", {})))

                span.set_attribute(f"{Config.LIBRARY_NAME}.span.type", "TOOL")
                span.set_attribute(f"{Config.LIBRARY_NAME}.entity.output", _serialize(block.content))

            except Exception as e:
                logger.error(f"Cannot set tool result attributes for tool_use_id={block.tool_use_id}: {e}")

    except Exception as e:
        logger.error(f"Error creating tool result span: {e}")


def set_request_attributes(span: Span, prompt: Any, options: ClaudeAgentOptions | None) -> int:
    """
    Write request metadata (model, system prompt, user prompt) to the root span.

    Args:
        span (Span): The root OpenTelemetry span to write attributes to.
        prompt (Any): The user prompt string for the request.
        options (ClaudeAgentOptions | None): Agent options containing model and system prompt.

    Returns:
        int: The next available prompt_index so callers can continue indexing result messages without gaps.
    """
    prompt_index = 0

    try:
        if options and isinstance(options, ClaudeAgentOptions):
            if model := options.model:
                span.set_attribute(SpanAttributes.LLM_REQUEST_MODEL, model)
            if system_prompt := options.system_prompt:
                prompt_index = _set_conversation(span, "system", system_prompt, prompt_index)
    except Exception as e:
        logger.error(f"Cannot extract options from request: {e}")

    try:
        if prompt and isinstance(prompt, str):
            prompt_index = _set_conversation(span, "user", prompt, prompt_index)
    except Exception as e:
        logger.error(f"Cannot extract prompt from request: {e}")

    return prompt_index


def set_system_message_attributes(span: Span, message: SystemMessage) -> None:
    """
    Write model info from a SystemMessage to the root span.

    Args:
        span (Span): The root OpenTelemetry span to write attributes to.
        message (SystemMessage): The system message containing model metadata.

    Returns:
        None
    """
    try:
        if model := message.data.get("model"):
            span.set_attribute(SpanAttributes.LLM_REQUEST_MODEL, model)
    except Exception as e:
        logger.error(f"Cannot extract model from SystemMessage: {e}")


def set_assistant_message_attributes(tracer: Tracer, parent_ctx: Context, message: AssistantMessage) -> None:
    """
    Create child spans for each content block in an AssistantMessage.

    Handles TextBlock, ThinkingBlock, and ToolUseBlock content. ToolUseBlocks are
    registered in the tool call registry for later correlation with their tool results.

    Args:
        tracer (Tracer): The OpenTelemetry tracer used to create child spans.
        parent_ctx (Context): The parent span context to attach child spans to.
        message (AssistantMessage): The assistant message containing one or more content blocks.

    Returns:
        None
    """
    for block in message.content:
        try:
            role, content, span_name = None, None, None

            if isinstance(block, TextBlock):
                role, span_name, content = "assistant", "claude-agent.assistant", block.text
            elif isinstance(block, ThinkingBlock):
                role, span_name, content = "assistant", "claude-agent.thinking", block.thinking
            elif isinstance(block, ToolUseBlock):
                with _tool_call_registry_lock:
                    _tool_call_registry[block.id] = {"name": block.name, "input": block.input}

            if role and content and span_name:
                with tracer.start_as_current_span(span_name, parent_ctx) as span:
                    if message.model:
                        span.set_attribute(SpanAttributes.LLM_REQUEST_MODEL, message.model)
                    _set_conversation(span, role, content)
        except Exception as e:
            logger.error(f"Cannot process assistant message block: {e}")


def set_user_message_attributes(tracer: Tracer, parent_ctx: Context, message: UserMessage) -> None:
    """
    Create a child tool-result span for each ToolResultBlock in a UserMessage.

    Args:
        tracer (Tracer): The OpenTelemetry tracer used to create child spans.
        parent_ctx (Context): The parent span context to attach child spans to.
        message (UserMessage): The user message containing one or more content blocks.

    Returns:
        None
    """
    for block in message.content:
        if isinstance(block, ToolResultBlock):
            _set_tool_result(tracer, parent_ctx, block)


def set_result_message_attributes(span: Span, message: ResultMessage, prompt_index: int = 0) -> None:
    """
    Write the final result text and token usage to the root span.

    Args:
        span (Span): The root OpenTelemetry span to write attributes to.
        message (ResultMessage): The result message containing the final text and usage data.
        prompt_index (int): The current prompt index for recording the result conversation. Defaults to 0.

    Returns:
        None
    """
    if result := message.result:
        _set_conversation(span, "assistant", result, prompt_index)

    if usage := message.usage:
        _set_usage(span, usage)



"""
Base evaluator class for Netra evaluation framework.

This module provides the abstract base class that all custom evaluators
should inherit from when implementing local evaluators for run_test_suite().
"""

from abc import ABC, abstractmethod

from netra.evaluation.models import EvaluatorConfig, EvaluatorContext, EvaluatorOutput


class BaseEvaluator(ABC):
    """
    Abstract base class for all evaluators.

    Subclasses must:
        - Implement evaluate(): Runs the evaluation logic and returns a result.

    The configuration is passed when instantiating the evaluator.

    Example:
        class MyEvaluator(BaseEvaluator):
            def evaluate(self, context: EvaluatorContext) -> EvaluatorOutput:
                is_correct = context.task_output == context.expected_output
                return EvaluatorOutput(
                    evaluator_name=self.config.name,
                    result=is_correct,
                    is_passed=is_correct,
                    reason="Output matches expected" if is_correct else "Mismatch",
                )

        # Usage:
        result = Netra.evaluation.run_test_suite(
            name="Copywriting Assistant v1",
            data=dataset,
            task=get_copywriting_agent_response,
            evaluators=[
                MyEvaluator(
                    EvaluatorConfig(
                        name="my_evaluator",
                        label="My Custom Evaluator",
                        score_type=ScoreType.BOOLEAN,
                    )
                )
            ]
        )
    """

    def __init__(self, config: EvaluatorConfig) -> None:
        """
        Initialize the evaluator with its configuration.

        Args:
            config: EvaluatorConfig containing:
                - name (str): Unique identifier for the evaluator.
                - label (str): Human-readable display name.
                - score_type (ScoreType): One of BOOLEAN, NUMERICAL, CATEGORICAL.
        """
        self.config = config

    @abstractmethod
    def evaluate(self, context: EvaluatorContext) -> EvaluatorOutput:
        """
        Run the evaluation logic.

        This method can be sync or async. If async, the framework will
        await the coroutine automatically.

        Args:
            context: EvaluatorContext containing:
                - input: The original input passed to the task function.
                - task_output: The output returned by the task function.
                - expected_output: The expected output from the dataset item.
                - metadata: Optional metadata from the dataset item.

        Returns:
            EvaluatorOutput: The evaluation result containing:
                - evaluator_name (str): Must match the name from config.
                - result (Any): The evaluation score/value.
                - is_passed (bool): Whether the evaluation passed.
                - reason (Optional[str]): Explanation for the result.
        """
